# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: auth_common.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11\x61uth_common.proto\x12\x18\x61rmonik.api.grpc.v1.auth\"<\n\x04User\x12\x10\n\x08username\x18\x01 \x01(\t\x12\r\n\x05roles\x18\x02 \x03(\t\x12\x13\n\x0bpermissions\x18\x03 \x03(\t\"\x17\n\x15GetCurrentUserRequest\"F\n\x16GetCurrentUserResponse\x12,\n\x04user\x18\x01 \x01(\x0b\x32\x1e.armonik.api.grpc.v1.auth.UserB=\n fr.aneo.armonik.api.grpc.v1.auth\xaa\x02\x18\x41rmoniK.Api.gRPC.V1.Authb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'auth_common_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n fr.aneo.armonik.api.grpc.v1.auth\252\002\030ArmoniK.Api.gRPC.V1.Auth'
  _globals['_USER']._serialized_start=47
  _globals['_USER']._serialized_end=107
  _globals['_GETCURRENTUSERREQUEST']._serialized_start=109
  _globals['_GETCURRENTUSERREQUEST']._serialized_end=132
  _globals['_GETCURRENTUSERRESPONSE']._serialized_start=134
  _globals['_GETCURRENTUSERRESPONSE']._serialized_end=204
# @@protoc_insertion_point(module_scope)



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor

class ResultStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    RESULT_STATUS_UNSPECIFIED: _ClassVar[ResultStatus]
    RESULT_STATUS_CREATED: _ClassVar[ResultStatus]
    RESULT_STATUS_COMPLETED: _ClassVar[ResultStatus]
    RESULT_STATUS_ABORTED: _ClassVar[ResultStatus]
    RESULT_STATUS_DELETED: _ClassVar[ResultStatus]
    RESULT_STATUS_NOTFOUND: _ClassVar[ResultStatus]
RESULT_STATUS_UNSPECIFIED: ResultStatus
RESULT_STATUS_CREATED: ResultStatus
RESULT_STATUS_COMPLETED: ResultStatus
RESULT_STATUS_ABORTED: ResultStatus
RESULT_STATUS_DELETED: ResultStatus
RESULT_STATUS_NOTFOUND: ResultStatus



from __future__ import annotations
from datetime import timedelta, datetime, timezone
from typing import List, Optional, Iterable, TypeVar

import google.protobuf.duration_pb2 as duration
import google.protobuf.timestamp_pb2 as timestamp

from ..protogen.common.submitter_common_pb2 import TaskFilter
from .enumwrapper import TaskStatus


T = TypeVar("T")


def get_task_filter(
    session_ids: Optional[List[str]] = None,
    task_ids: Optional[List[str]] = None,
    included_statuses: Optional[List[TaskStatus]] = None,
    excluded_statuses: Optional[List[TaskStatus]] = None,
) -> TaskFilter:
    """Helper function to generate a task filter from the parameters

    Args:
        session_ids: Optional list of session Ids to filter against, mutually exclusive with task_ids
        task_ids: Optional list of task ids to filter against, mutually exclusive with session_ids
        included_statuses: Optional list of task statuses to include in the filter, mutually exclusive with excluded_statuses
        excluded_statuses: Optional list of task statuses to exclude in the filter, mutually exclusive with included_statuses

    Returns:
        Task filter to be used in a gRPC call to filter tasks
    """
    if session_ids and task_ids:
        raise ValueError("session_ids and task_ids cannot be defined at the same time")
    if included_statuses and excluded_statuses:
        raise ValueError(
            "included_statuses and excluded_statuses cannot be defined at the same time"
        )
    task_filter = TaskFilter(
        session=TaskFilter.IdsRequest() if session_ids else None,
        task=TaskFilter.IdsRequest() if task_ids else None,
        included=TaskFilter.StatusesRequest() if included_statuses else None,
        excluded=TaskFilter.StatusesRequest() if excluded_statuses else None,
    )
    if session_ids:
        task_filter.session.ids.extend(session_ids)
    if task_ids:
        task_filter.task.ids.extend(task_ids)
    if included_statuses:
        task_filter.included.statuses.extend(included_statuses)
    if excluded_statuses:
        task_filter.excluded.statuses.extend(excluded_statuses)
    return task_filter


def datetime_to_timestamp(time_stamp: Optional[datetime]) -> timestamp.Timestamp:
    """Helper function to convert a Python Datetime to a gRPC Timestamp

    Args:
        time_stamp: Python datetime timestamp to convert

    Returns:
        Equivalent gRPC Timestamp
    """
    t = timestamp.Timestamp()
    t.FromDatetime(
        dt=time_stamp if time_stamp is not None else datetime(1970, 1, 1, 0, 0, tzinfo=timezone.utc)
    )
    return t


def timestamp_to_datetime(time_stamp: timestamp.Timestamp) -> Optional[datetime]:
    """Helper function to convert a gRPC Timestamp to a Python Datetime
    Note that datetime has microseconds accuracy instead of nanosecond accuracy for gRPC Timestamp
    Therefore, the conversion may not be lossless.
    Args:
        time_stamp: gRPC Timestamp to convert

    Returns:
        Equivalent Python Datetime
    """
    if time_stamp.seconds == 0 and time_stamp.nanos == 0:
        return
    return time_stamp.ToDatetime(tzinfo=timezone.utc)


def duration_to_timedelta(delta: duration.Duration) -> timedelta:
    """Helper function to convert a gRPC Duration into a Python timedelta
    Note that timedelta has microseconds accuracy instead of nanosecond accuracy for gRPC Duration.
    Therefore, the conversion may not be lossless.
    Args:
        delta: gRPC Duration to convert

    Returns:
        Equivalent Python timedelta
    """
    return delta.ToTimedelta()


def timedelta_to_duration(delta: timedelta) -> duration.Duration:
    """Helper function to convert a gRPC Duration to a Python Datetime

    Args:
        delta: Python timedelta to convert

    Returns:
        Equivalent gRPC Duration
    """
    d = duration.Duration()
    d.FromTimedelta(delta)
    return d


def batched(iterable: Iterable[T], n: int) -> Iterable[List[T]]:
    """
    Batches elements from an iterable into lists of size at most 'n'.

    Args:
        iterable : The input iterable.
        n : The batch size.

    Yields:
        A generator yielding batches of elements from the input iterable.
    """
    it = iter(iterable)

    sentinel = object()
    batch = []
    c = next(it, sentinel)
    while c is not sentinel:
        batch.append(c)
        if len(batch) == n:
            yield batch
            batch = []
        c = next(it, sentinel)
    if len(batch) > 0:
        yield batch



from __future__ import annotations

import json
from abc import abstractmethod, ABC
from datetime import datetime, timedelta
from enum import Enum, auto
from typing import Optional, Union, Any, List, Dict, Generic, TypeVar, overload, Type, Tuple

# noinspection PyUnresolvedReferences
from google.protobuf.duration_pb2 import Duration
from google.protobuf.message import Message

# noinspection PyUnresolvedReferences
from google.protobuf.timestamp_pb2 import Timestamp

from ._message_types import (
    BasicMessageType,
    InnerMessageType,
    SimpleOperator,
    DisjunctionType,
    ConjunctionType,
)
from ..helpers import datetime_to_timestamp, timedelta_to_duration
from ...protogen.common.filters_common_pb2 import (
    FILTER_STRING_OPERATOR_EQUAL,
    FILTER_STRING_OPERATOR_NOT_EQUAL,
    FILTER_STRING_OPERATOR_CONTAINS,
    FILTER_STRING_OPERATOR_NOT_CONTAINS,
    FilterString,
    FILTER_STRING_OPERATOR_STARTS_WITH,
    FILTER_STRING_OPERATOR_ENDS_WITH,
    FILTER_STATUS_OPERATOR_EQUAL,
    FILTER_STATUS_OPERATOR_NOT_EQUAL,
    FILTER_DATE_OPERATOR_EQUAL,
    FILTER_DATE_OPERATOR_NOT_EQUAL,
    FILTER_DATE_OPERATOR_BEFORE,
    FILTER_DATE_OPERATOR_BEFORE_OR_EQUAL,
    FILTER_DATE_OPERATOR_AFTER,
    FILTER_DATE_OPERATOR_AFTER_OR_EQUAL,
    FILTER_NUMBER_OPERATOR_EQUAL,
    FILTER_NUMBER_OPERATOR_NOT_EQUAL,
    FILTER_NUMBER_OPERATOR_LESS_THAN,
    FILTER_NUMBER_OPERATOR_LESS_THAN_OR_EQUAL,
    FILTER_NUMBER_OPERATOR_GREATER_THAN,
    FILTER_NUMBER_OPERATOR_GREATER_THAN_OR_EQUAL,
    FilterDate,
    FilterNumber,
    FILTER_BOOLEAN_OPERATOR_IS,
    FilterBoolean,
    FILTER_ARRAY_OPERATOR_CONTAINS,
    FILTER_ARRAY_OPERATOR_NOT_CONTAINS,
    FilterArray,
    FILTER_DURATION_OPERATOR_EQUAL,
    FILTER_DURATION_OPERATOR_NOT_EQUAL,
    FILTER_DURATION_OPERATOR_SHORTER_THAN,
    FILTER_DURATION_OPERATOR_SHORTER_THAN_OR_EQUAL,
    FILTER_DURATION_OPERATOR_LONGER_THAN,
    FILTER_DURATION_OPERATOR_LONGER_THAN_OR_EQUAL,
    FilterDuration,
)


class FilterError(ValueError):
    """
    Exception raised when there is an error related to filter
    """

    def __init__(self, filter_instance: Optional[Filter], message: str):
        """

        Args:
            filter_instance: Instance of the filter
            message: Accompanying message
        """
        self.filter = filter_instance
        self.message = message

    def __str__(self):
        return f"Filter: {str(self.filter)} Error: {self.message}"


eq_special = object()
ne_special = object()


class Filter:
    """
    Base class of all filters.
    How to use it :
    A filter should not be created directly by the end user. Instead, users should use the class static properties to indicate what filter they want to use.
    For example, if a user wants to list tasks that are in session "xxx", then they should use :
    Task.session_id == "xxx"
    Users can combine filters using the binary operators (& | ~).

    When combining filters together, the new filter combination is always kept in the disjunctive normal form to be compatible with ArmoniK.
    """

    eq_ = eq_special  # Special case
    ne_ = ne_special  # Special case
    lt_ = None
    le_ = None
    gt_ = None
    ge_ = None
    contains_ = None
    notcontains_ = None
    value_type_ = None
    inner_message_type_attr_ = None

    def __init__(
        self,
        field: Optional[Message],
        disjunction_message_type: DisjunctionType,
        conjunction_message_type: ConjunctionType,
        message_type: Union[BasicMessageType, DisjunctionType],
        inner_message_type: Optional[InnerMessageType],
        operator: Union[SimpleOperator, None] = None,
        value: Any = None,
    ):
        """
        Init a new filter
        Notes: This class should not be instanced directly. Always use the proxies in armonik.common.filter
        Args:
            field: gRPC Message field
            disjunction_message_type: Message type of the logical OR
            conjunction_message_type: Message type of the logical AND
            message_type: Filter gRPC message type
            inner_message_type: Filter type gRPC message type
            operator: Filter operator
            value: Value of the filter
        """
        self.field = field
        self.disjunction_message_type = disjunction_message_type
        self.conjunction_message_type = conjunction_message_type
        self.message_type = message_type
        self.inner_message_type = inner_message_type
        self.operator = operator
        self.value = value

    def _disjunction(self, filters: List[List[Filter]]) -> Filter:
        """
        Creates a disjunction from the current Filter types and a list of lists of filter
        Args:
            filters: List of Lists of Filters (OR of ANDs)

        Returns:
            New Filter instance as disjunction
        """
        return Filter(
            None,
            self.disjunction_message_type,
            self.conjunction_message_type,
            self.disjunction_message_type,
            None,
            None,
            filters,
        )

    def _check_compatible(self, other: Filter):
        """
        Checks if the given Filter is compatible with the current Filter. Raises ValueError if they are not compatible.
        Args:
            other: Other Filter

        Returns:
            None
        """
        if not isinstance(other, Filter):
            msg = f"{str(self)} is not compatible with {str(other)} as the latter is not a Filter"
            raise ValueError(msg)
        if not (
            self.disjunction_message_type == other.disjunction_message_type
            and self.conjunction_message_type == other.conjunction_message_type
        ):
            msg = (
                f"{str(self)} is not compatible with "
                f"{str(other)} as they apply to different object types"
            )
            raise ValueError(msg)

    def _is_empty(self) -> bool:
        """
        Checks if the current Filter is empty
        Returns:
            True if the current Filter is empty
        """
        return self.value is None or (
            isinstance(self.value, list)
            and (
                len(self.value) == 0
                or (isinstance(self.value[0], list) and len(self.value[0]) == 0)
            )
        )

    def _is_disjunction(self) -> bool:
        """
        Checks if the current Filter is a disjunction
        Returns:
            True if the current Filter is a disjunction
        """
        return self.message_type == self.disjunction_message_type

    def to_disjunction(self) -> Filter:
        """
        Wraps the current Filter into a disjunction. Does nothing if the current Filter is already a disjunction.
        Returns:
            Disjunction Filter
        Raises:
            FilterError if the filter is empty
        """
        if self._is_empty():
            raise FilterError(self, "Empty filter")
        if isinstance(self.operator, SimpleOperator):
            return self._disjunction([[self]])
        return self

    def __and__(self, other: Filter) -> Filter:
        """
        Logical and
        """
        self._check_compatible(other)
        # (a | (b&c)) & (d | (e&f)) = a&d | a&e&f | b&c&d | b&c&e&f
        return self._disjunction(
            [c1 + c2 for c2 in other.to_disjunction().value for c1 in self.to_disjunction().value]
        )

    def __mul__(self, other: Filter) -> Filter:
        """
        Same as A & B
        """
        return self & other

    def __or__(self, other: Filter) -> Filter:
        """
        Logical or
        """
        self._check_compatible(other)
        # (a | b&c) | (d | e&f) = a | b&c | d | e&f
        return self._disjunction(self.to_disjunction().value + other.to_disjunction().value)

    def __add__(self, other: Filter) -> Filter:
        """
        Same as Logical or
        """
        return self | other

    def __eq__(self, value) -> Filter:
        """
        Creates a Filter which checks for the equality of the property compared to the value
        Args:
            value: Value to check against

        Returns:
            New Filter instance
        """
        return self._change_operation(value, self.__class__.eq_, "==")

    def __ne__(self, value) -> Filter:
        """
        Creates a Filter which checks for the inequality of the property compared to the value
        Args:
            value: Value to check against

        Returns:
            New Filter instance
        """
        return self._change_operation(value, self.__class__.ne_, "!=")

    def __lt__(self, value) -> Filter:
        """
        Creates a Filter which checks that the object's value is less than the value
        Args:
            value: Value to check against

        Returns:
            New Filter instance
        """
        return self._change_operation(value, self.__class__.lt_, "<")

    def __le__(self, value) -> Filter:
        """
        Creates a Filter which checks that the object's value is less than or equal to the value
        Args:
            value: Value to check against

        Returns:
            New Filter instance
        """
        return self._change_operation(value, self.__class__.le_, "<=")

    def __gt__(self, value) -> Filter:
        """
        Creates a Filter which checks that the object's value is greater than to the value
        Args:
            value: Value to check against

        Returns:
            New Filter instance
        """
        return self._change_operation(value, self.__class__.gt_, ">")

    def __ge__(self, value) -> Filter:
        """
        Creates a Filter which checks that the object's value is greater than or equal to the value
        Args:
            value: Value to check against

        Returns:
            New Filter instance
        """
        return self._change_operation(value, self.__class__.ge_, ">=")

    def contains(self, value) -> Filter:
        """
        Creates a Filter which checks that the object's value contains the value
        Args:
            value: Value to check against

        Returns:
            New Filter instance
        """
        return self._change_operation(value, self.__class__.contains_, "contains")

    def __invert__(self) -> Filter:
        """
        Logical not
        """
        if self.operator is None:
            # No operator
            if self._is_empty():
                raise FilterError(self, "Cannot invert filter without a value or operator")
            try:
                # The filter is a combination
                # ~(a | (b & c)) => ~a & (~b | ~c) => ~a&~b | ~a&~c
                new_filter = self._disjunction([[~c] for c in self.value[0]])
                for conj in self.value[1:]:
                    new_filter &= self._disjunction([[~c] for c in conj])
                return new_filter
            except FilterError as e:
                msg = f"Could not invert because of an error in subfilter inversion : {str(e)}"
                raise FilterError(self, msg)
        # Invert the operation
        if self.operator == self.__class__.eq_:
            return self._change_operation(self.value, self.__class__.ne_, "!=", True)
        if self.operator == self.__class__.ne_:
            return self._change_operation(self.value, self.__class__.eq_, "==", True)
        if self.operator == self.__class__.lt_:
            return self._change_operation(self.value, self.__class__.ge_, ">=", True)
        if self.operator == self.__class__.le_:
            return self._change_operation(self.value, self.__class__.gt_, ">", True)
        if self.operator == self.__class__.gt_:
            return self._change_operation(self.value, self.__class__.le_, "<=", True)
        if self.operator == self.__class__.ge_:
            return self._change_operation(self.value, self.__class__.lt_, "<", True)
        if self.operator == self.__class__.contains_:
            return self._change_operation(
                self.value, self.__class__.notcontains_, "not_contains", True
            )
        if self.operator == self.__class__.notcontains_:
            return self._change_operation(self.value, self.__class__.contains_, "contains", True)
        msg = f"{self.__class__.__name__} operator {str(self.operator)} for field {str(self.field)} has no inverted equivalent"
        raise FilterError(self, msg)

    def __neg__(self) -> Filter:
        """
        Same as Logical not
        """
        return ~self

    def __xor__(self, other: Filter) -> Filter:
        """
        Logical xor
        """
        return (self & (~other)) | ((~self) & other)

    def _sanitize_value(self, value: Any) -> Any:
        """
        Takes an input value and sanitizes it to be compatible with the filter.
        If the type is incompatible, throws a FilterError
        Args:
            value: Value to be sanitized

        Returns:
            Sanitized value
        """
        if self.__class__.value_type_ is None or isinstance(value, self.__class__.value_type_):
            return value
        msg = f"Expected value type '{self.__class__.value_type_.__name__}' for field '{self.field_name}', got '{value.__class__.__name__}' instead"
        raise FilterError(None, msg)

    def _change_operation(
        self, value: Any, operator: Any, operator_str: str = "", override: bool = False
    ) -> Filter:
        """
        Internal function to create a new filter from the current filter with a different value and/or operator
        Args:
            value: Value of the new filter
            operator: Operator of the new filter
            operator_str: Optional string for error message clarification
            override: If set to true, bypasses the check for reassignment
        Returns:
            new filter with the given value/operator

        Raises:
            FilterError if the given operator is not available for the given class
        """
        if not override:
            if self.value is not None:
                if isinstance(value, bool):
                    if operator is self.__class__.eq_ and self.__class__.eq_ is not None:
                        return self if value else ~self
                    if operator is self.__class__.ne_ and self.__class__.ne_ is not None:
                        return ~self if value else self
                if self._is_disjunction():
                    raise FilterError(self, "Cannot apply operator to a filter combination")
                raise FilterError(self, "Cannot apply operator to an already defined filter")
        if operator is None or self._is_disjunction():
            msg = f"Operator {operator_str} is not available for {self.__class__.__name__}."
            raise FilterError(self, msg)
        return self.__class__(
            self.field,
            self.disjunction_message_type,
            self.conjunction_message_type,
            self.message_type,
            self.inner_message_type,
            operator,
            self._sanitize_value(value),
        )

    def to_dict(self) -> Dict[str, Any]:
        """
        Transforms the filter in a dictionary describing it
        Returns:
            Dictionary representing the filter in a disjunctive normal form
        """
        if not self._is_disjunction() or self.value is None:
            return {
                "field": str(self.field),
                "value": str(self.value),
                "operator": str(self.operator),
            }
        if len(self.value) > 1:
            return {
                "or": [
                    ({"and": [c.to_dict() for c in conj]} if len(conj) > 1 else conj[0].to_dict())
                    for conj in self.value
                ]
            }
        return (
            (
                {"and": [c.to_dict() for c in self.value[0]]}
                if len(self.value[0]) > 1
                else self.value[0][0].to_dict()
            )
            if not self._is_empty()
            else {}
        )

    @property
    def field_name(self) -> str:
        return str(self.field).replace(" ", "").split("\n")[1].split("ENUM_FIELD_")[1].lower()

    def __str__(self) -> str:
        return json.dumps(self.to_dict())

    def to_basic_message(self) -> BasicMessageType:
        """
        Converts the filter into its base gRPC message
        Returns:
            Base gRPC message
        """
        if self.__class__.inner_message_type_attr_ is None:
            raise FilterError(
                self,
                "Cannot use an empty filter in combination with others. "
                "Did you forget to write the condition ?",
            )
        return self.message_type(
            field=self.field,
            **{
                self.__class__.inner_message_type_attr_: self.inner_message_type(
                    value=self.value, operator=self.operator
                )
            },
        )

    def to_message(self) -> DisjunctionType:
        """
        Converts the filter into its gRPC disjunctive message
        Returns:
            Disjunctive gRPC message
        """
        if not self._is_disjunction():
            # Convert the message to a disjunction if it's not the case
            return self.to_disjunction().to_message()

        def to_conjunction_message(conj: List[Filter]) -> ConjunctionType:
            conj_raw = self.conjunction_message_type()
            # Need to use getattr because and is a reserved name
            getattr(conj_raw, "and").extend(f.to_basic_message() for f in conj)
            return conj_raw

        raw = self.disjunction_message_type()
        # Need to use getattr because or is a reserved name
        getattr(raw, "or").extend(
            (to_conjunction_message(conj) for conj in self.value) if self.value is not None else []
        )
        return raw

    def __bool__(self):
        raise FilterError(
            self,
            "Filters cannot be transformed into booleans. "
            "You may see this error if you try to combine filters with 'or', 'and', 'in', or 'not'. "
            "Use '|', '&', '.contains' or '~' instead.",
        )


class StringFilter(Filter):
    """
    Filter for string comparisons
    """

    eq_ = FILTER_STRING_OPERATOR_EQUAL
    ne_ = FILTER_STRING_OPERATOR_NOT_EQUAL
    contains_ = FILTER_STRING_OPERATOR_CONTAINS
    notcontains_ = FILTER_STRING_OPERATOR_NOT_CONTAINS
    value_type_ = str
    inner_message_type_attr_ = "filter_string"

    def __init__(
        self,
        field: Optional[Message],
        disjunction_message_type: DisjunctionType,
        conjunction_message_type: ConjunctionType,
        message_type: Union[BasicMessageType, DisjunctionType],
        inner_message_type: Optional[InnerMessageType] = FilterString,
        operator: Union[SimpleOperator, None] = None,
        value: Any = None,
    ):
        super().__init__(
            field,
            disjunction_message_type,
            conjunction_message_type,
            message_type,
            inner_message_type,
            operator,
            value,
        )

    def startswith(self, value: str) -> "StringFilter":
        """
        Creates a Filter which checks that the object's value starts with the value
        Args:
            value: Value to check against

        Returns:
            New Filter instance
        """
        return self._change_operation(value, FILTER_STRING_OPERATOR_STARTS_WITH, "startswith")

    def endswith(self, value: str) -> "StringFilter":
        """
        Creates a Filter which checks that the object's value ends with the value
        Args:
            value: Value to check against

        Returns:
            New Filter instance
        """
        return self._change_operation(value, FILTER_STRING_OPERATOR_ENDS_WITH, "endswith")

    def __repr__(self) -> str:
        return f'{str(self.field)} {str(self.operator)} "{str(self.value)}"'


class StatusFilter(Filter):
    """
    Filter for status comparison
    """

    eq_ = FILTER_STATUS_OPERATOR_EQUAL
    ne_ = FILTER_STATUS_OPERATOR_NOT_EQUAL
    inner_message_type_attr_ = "filter_status"

    def __init__(
        self,
        field: Optional[Message],
        disjunction_message_type: DisjunctionType,
        conjunction_message_type: ConjunctionType,
        message_type: Union[BasicMessageType, DisjunctionType],
        inner_message_type: InnerMessageType,
        operator: Union[SimpleOperator, None] = None,
        value: Any = None,
    ):
        super().__init__(
            field,
            disjunction_message_type,
            conjunction_message_type,
            message_type,
            inner_message_type,
            operator,
            value,
        )


class DateFilter(Filter):
    """
    Filter for timestamp comparison
    """

    eq_ = FILTER_DATE_OPERATOR_EQUAL
    ne_ = FILTER_DATE_OPERATOR_NOT_EQUAL
    lt_ = FILTER_DATE_OPERATOR_BEFORE
    le_ = FILTER_DATE_OPERATOR_BEFORE_OR_EQUAL
    gt_ = FILTER_DATE_OPERATOR_AFTER
    ge_ = FILTER_DATE_OPERATOR_AFTER_OR_EQUAL
    inner_message_type_attr_ = "filter_date"

    def __init__(
        self,
        field: Optional[Message],
        disjunction_message_type: DisjunctionType,
        conjunction_message_type: ConjunctionType,
        message_type: Union[BasicMessageType, DisjunctionType],
        inner_message_type: InnerMessageType = FilterDate,
        operator: Union[SimpleOperator, None] = None,
        value: Any = None,
    ):
        super().__init__(
            field,
            disjunction_message_type,
            conjunction_message_type,
            message_type,
            inner_message_type,
            operator,
            value,
        )

    def _sanitize_value(self, value: Any) -> Any:
        if isinstance(value, datetime):
            return datetime_to_timestamp(value)
        if isinstance(value, Timestamp):
            return value
        msg = (
            f"Expected value type '{datetime.__name__}' or '{Timestamp.__name__}' "
            f"for field '{self.field_name}', got '{type(value).__name__}' instead"
        )
        raise FilterError(self, msg)


class NumberFilter(Filter):
    """
    Filter for int comparison
    """

    eq_ = FILTER_NUMBER_OPERATOR_EQUAL
    ne_ = FILTER_NUMBER_OPERATOR_NOT_EQUAL
    lt_ = FILTER_NUMBER_OPERATOR_LESS_THAN
    le_ = FILTER_NUMBER_OPERATOR_LESS_THAN_OR_EQUAL
    gt_ = FILTER_NUMBER_OPERATOR_GREATER_THAN
    ge_ = FILTER_NUMBER_OPERATOR_GREATER_THAN_OR_EQUAL
    value_type_ = int
    inner_message_type_attr_ = "filter_number"

    def __init__(
        self,
        field: Optional[Message],
        disjunction_message_type: DisjunctionType,
        conjunction_message_type: ConjunctionType,
        message_type: Union[BasicMessageType, DisjunctionType],
        inner_message_type: InnerMessageType = FilterNumber,
        operator: Union[SimpleOperator, None] = None,
        value: Any = None,
    ):
        super().__init__(
            field,
            disjunction_message_type,
            conjunction_message_type,
            message_type,
            inner_message_type,
            operator,
            value,
        )


class BooleanFilter(Filter):
    """
    Filter for boolean comparison
    """

    eq_ = FILTER_BOOLEAN_OPERATOR_IS
    value_type_ = bool
    inner_message_type_attr_ = "filter_boolean"

    def __init__(
        self,
        field: Optional[Message],
        disjunction_message_type: DisjunctionType,
        conjunction_message_type: ConjunctionType,
        message_type: Union[BasicMessageType, DisjunctionType],
        inner_message_type: InnerMessageType = FilterBoolean,
        operator: Union[SimpleOperator, None] = FILTER_BOOLEAN_OPERATOR_IS,
        value: Any = True,
    ):
        super().__init__(
            field,
            disjunction_message_type,
            conjunction_message_type,
            message_type,
            inner_message_type,
            operator,
            value,
        )

    def __ne__(self, value: bool) -> BooleanFilter:
        return self._change_operation(not value, self.__class__.eq_, "!=", True)

    def __invert__(self) -> BooleanFilter:
        return self._change_operation(not self.value, self.__class__.eq_, "==", True)


class ArrayFilter(Filter):
    """
    Filter for array comparisons
    """

    contains_ = FILTER_ARRAY_OPERATOR_CONTAINS
    notcontains_ = FILTER_ARRAY_OPERATOR_NOT_CONTAINS
    value_type_ = str
    inner_message_type_attr_ = "filter_array"

    def __init__(
        self,
        field: Optional[Message],
        disjunction_message_type: DisjunctionType,
        conjunction_message_type: ConjunctionType,
        message_type: Union[BasicMessageType, DisjunctionType],
        inner_message_type: InnerMessageType = FilterArray,
        operator: Union[SimpleOperator, None] = None,
        value: Any = None,
    ):
        super().__init__(
            field,
            disjunction_message_type,
            conjunction_message_type,
            message_type,
            inner_message_type,
            operator,
            value,
        )


class DurationFilter(Filter):
    """
    Filter for duration comparison
    """

    eq_ = FILTER_DURATION_OPERATOR_EQUAL
    ne_ = FILTER_DURATION_OPERATOR_NOT_EQUAL
    lt_ = FILTER_DURATION_OPERATOR_SHORTER_THAN
    le_ = FILTER_DURATION_OPERATOR_SHORTER_THAN_OR_EQUAL
    gt_ = FILTER_DURATION_OPERATOR_LONGER_THAN
    ge_ = FILTER_DURATION_OPERATOR_LONGER_THAN_OR_EQUAL
    value_type_ = Duration
    inner_message_type_attr_ = "filter_duration"

    def __init__(
        self,
        field: Optional[Message],
        disjunction_message_type: DisjunctionType,
        conjunction_message_type: ConjunctionType,
        message_type: Union[BasicMessageType, DisjunctionType],
        inner_message_type: InnerMessageType = FilterDuration,
        operator: Union[SimpleOperator, None] = None,
        value: Any = None,
    ):
        super().__init__(
            field,
            disjunction_message_type,
            conjunction_message_type,
            message_type,
            inner_message_type,
            operator,
            value,
        )

    def _sanitize_value(self, value: Any) -> Any:
        if isinstance(value, timedelta):
            return timedelta_to_duration(value)
        if isinstance(value, Duration):
            return value
        msg = (
            f"Expected value type '{timedelta.__name__}' or '{Duration.__name__}' "
            f"for field '{self.field_name}', got '{type(value).__name__}' instead."
        )
        raise FilterError(self, msg)


class FType(Enum):
    """
    Enum for the type of a filter. Used internally.
    """

    UNKNOWN = auto()
    NA = auto()
    NUM = auto()
    STR = auto()
    ARRAY = auto()
    DURATION = auto()
    DATE = auto()
    STATUS = auto()
    BOOL = auto()


def _raise(field):
    """
    The field is unknown in the class
    Args:
        field: Unknown field

    Returns:

    """
    msg = f"Unknown field {field}"
    raise ValueError(msg)


def _na(field):
    msg = f"Field {field} is not available as a filter"
    raise FilterError(None, msg)


class FilterWrapper(Filter, ABC):
    """
    Wraps the filter creation to alleviate repetitions
    """

    _fields: Dict[str, Tuple[FType, Optional[Any]]] = {}

    def __init__(
        self,
        disjunction_message_type: DisjunctionType,
        conjunction_message_type: ConjunctionType,
        message_type: BasicMessageType,
        status_type: Optional[InnerMessageType] = None,
    ):
        super().__init__(
            None,
            disjunction_message_type,
            conjunction_message_type,
            disjunction_message_type,
            status_type,
        )
        self._vtable = {
            FType.UNKNOWN: _raise,
            FType.NA: _na,
            FType.NUM: self._number,
            FType.STR: self._string,
            FType.ARRAY: self._array,
            FType.DURATION: self._duration,
            FType.DATE: self._date,
            FType.STATUS: self._status,
            FType.BOOL: self._bool,
        }
        self.basic_message_type = message_type

    def _string(self, field: Message) -> StringFilter:
        """
        Creates a new string filter on the given field
        Args:
            field: field to filter against

        Returns:
            New string filter on the given field
        """
        return StringFilter(
            field,
            self.disjunction_message_type,
            self.conjunction_message_type,
            self.basic_message_type,
        )

    def _bool(self, field: Message) -> BooleanFilter:
        """
        Creates a new boolean filter on the given field
        Args:
            field: field to filter against

        Returns:
            New boolean filter on the given field
        """
        return BooleanFilter(
            field,
            self.disjunction_message_type,
            self.conjunction_message_type,
            self.basic_message_type,
        )

    def _array(self, field: Message) -> ArrayFilter:
        """
        Creates a new array filter on the given field
        Args:
            field: field to filter against

        Returns:
            New array filter on the given field
        """
        return ArrayFilter(
            field,
            self.disjunction_message_type,
            self.conjunction_message_type,
            self.basic_message_type,
        )

    def _duration(self, field: Message) -> DurationFilter:
        """
        Creates a new duration filter on the given field
        Args:
            field: field to filter against

        Returns:
            New duration filter on the given field

        """
        return DurationFilter(
            field,
            self.disjunction_message_type,
            self.conjunction_message_type,
            self.basic_message_type,
        )

    def _date(self, field: Message) -> DateFilter:
        """
        Creates a new date filter on the given field
        Args:
            field:  field to filter against

        Returns:
            New date filter on the given field

        """
        return DateFilter(
            field,
            self.disjunction_message_type,
            self.conjunction_message_type,
            self.basic_message_type,
        )

    def _number(self, field: Message) -> NumberFilter:
        """
        Creates a new number filter on the given field
        Args:
            field:  field to filter against

        Returns:
            New number filter on the given field

        """
        return NumberFilter(
            field,
            self.disjunction_message_type,
            self.conjunction_message_type,
            self.basic_message_type,
        )

    def _status(self, field: Message) -> StatusFilter:
        """
        Creates a new status filter on the given field
        Args:
            field:  field to filter against

        Returns:
            New status filter on the given field

        """
        return StatusFilter(
            field,
            self.disjunction_message_type,
            self.conjunction_message_type,
            self.basic_message_type,
            self.inner_message_type,
        )

    @abstractmethod
    def _build_field(self, field: Any) -> Message:
        """
        Build the gRPC field object
        Args:
            field: base field enum value

        Returns:
            Field message
        """
        ...

    def __call__(self, field_name: str) -> Filter:
        """
        Get the Filter associated with the field name
        Args:
            field_name: Field name

        Returns:
            Filter associated with the field name
        """
        ftype, field_value = self.__class__._fields.get(field_name, (FType.UNKNOWN, field_name))
        return self._vtable[ftype](
            self._build_field(field_value)
            if ftype not in (FType.UNKNOWN, FType.NA)
            else field_value
        )


AttributeType = TypeVar("AttributeType")


class FilterDescriptor(Generic[AttributeType]):
    """
    Descriptor used to declare a filter
    """

    def __init__(self, wrapper: FilterWrapper):
        self.wrapper = wrapper

    def __set_name__(self, owner: Type, name: str):
        self.name = name
        self._name = "_" + name

    @overload
    def __get__(self, instance: None, owner: Type) -> Filter: ...

    @overload
    def __get__(self, instance: object, owner: Type) -> AttributeType: ...

    def __get__(
        self, instance: Optional[object], owner: Optional[Type]
    ) -> Union[Filter, AttributeType]:
        # When called statically, returns the filter
        if instance is None:
            return self.wrapper(self.name)
        # Otherwise gets the object's hidden property
        return getattr(instance, self._name)

    def __set__(self, instance: object, value: AttributeType):
        setattr(instance, self._name, value)



from __future__ import annotations

from typing import Dict, List, Tuple, cast, Optional, Union, Iterable

from deprecation import deprecated
from grpc import Channel

from .. import __version__
from ..common import Direction, Result
from ..common.filter import Filter, ResultFilter
from ..common.helpers import batched
from ..protogen.client.results_service_pb2_grpc import ResultsStub
from ..protogen.common.objects_pb2 import Empty
from ..protogen.common.results_common_pb2 import (
    CreateResultsMetaDataRequest,
    CreateResultsMetaDataResponse,
    CreateResultsRequest,
    CreateResultsResponse,
    DeleteResultsDataRequest,
    DownloadResultDataRequest,
    GetOwnerTaskIdRequest,
    GetOwnerTaskIdResponse,
    GetResultRequest,
    GetResultResponse,
    ListResultsRequest,
    ListResultsResponse,
    ResultsServiceConfigurationResponse,
    UploadResultDataRequest,
    ImportResultsDataRequest,
    ImportResultsDataResponse,
)
from ..protogen.common.results_fields_pb2 import ResultField
from ..protogen.common.sort_direction_pb2 import SortDirection


@deprecated("3.19.0", None, __version__, "Use Result.<name of the field> instead")
class ResultFieldFilter:
    STATUS = Result.status
    RESULT_ID = Result.result_id


class ArmoniKResults:
    def __init__(self, grpc_channel: Channel):
        """Result service client

        Args:
            grpc_channel: gRPC channel to use
        """
        self._client = ResultsStub(grpc_channel)

    @deprecated(
        deprecated_in="3.15.0",
        details="Use create_result_metadata or create_result insted.",
    )
    def get_results_ids(self, session_id: str, names: List[str]) -> Dict[str, str]:
        return {
            r.name: r.result_id
            for r in cast(
                CreateResultsMetaDataResponse,
                self._client.CreateResultsMetaData(
                    CreateResultsMetaDataRequest(
                        results=[CreateResultsMetaDataRequest.ResultCreate(name=n) for n in names],
                        session_id=session_id,
                    )
                ),
            ).results
        }

    def list_results(
        self,
        result_filter: Optional[Filter] = None,
        page: int = 0,
        page_size: int = 1000,
        sort_field: Filter = Result.status,
        sort_direction: SortDirection = Direction.ASC,
    ) -> Tuple[int, List[Result]]:
        """List results based on a filter.

        Args:
            result_filter (Filter): Filter to apply when listing results
            page: page number to request, useful for pagination, defaults to 0
            page_size: size of a page, defaults to 1000
            sort_field: field to sort the resulting list by, defaults to the status
            sort_direction: direction of the sort, defaults to ascending
        Returns:
            A tuple containing :
            - The total number of results for the given filter
            - The obtained list of results
        """
        request: ListResultsRequest = ListResultsRequest(
            page=page,
            page_size=page_size,
            filters=(
                ResultFilter().to_message() if result_filter is None else result_filter.to_message()
            ),
            sort=ListResultsRequest.Sort(
                field=cast(ResultField, sort_field.field), direction=sort_direction
            ),
        )
        list_response: ListResultsResponse = self._client.ListResults(request)
        return list_response.total, [Result.from_message(r) for r in list_response.results]

    def get_result(self, result_id: str) -> Result:
        """Get a result by id.

        Args:
            result_id: The ID of the result.

        Return:
            The result summary.
        """
        request = GetResultRequest(result_id=result_id)
        response: GetResultResponse = self._client.GetResult(request)
        return Result.from_message(response.result)

    def get_owner_task_id(
        self, result_ids: List[str], session_id: str, batch_size: int = 500
    ) -> Dict[str, str]:
        """Get the IDs of the tasks that should produce the results.

        Args:
            result_ids: A list of results.
            session_id: The ID of the session to which the results belongs.
            batch_size: Batch size for querying.

        Return:
            A dictionnary mapping results to owner task ID.
        """
        results = {}
        for result_ids_batch in batched(result_ids, batch_size):
            request = GetOwnerTaskIdRequest(session_id=session_id, result_id=result_ids_batch)
            response: GetOwnerTaskIdResponse = self._client.GetOwnerTaskId(request)
            for result_task in response.result_task:
                results[result_task.result_id] = result_task.task_id
        return results

    def create_results_metadata(
        self, result_names: List[str], session_id: str, batch_size: int = 100
    ) -> Dict[str, Result]:
        """Create the metadata of multiple results at once.
        Data have to be uploaded separately.

        Args:
            result_names: The list of the names of the results to create.
            session_id: The ID of the session to which the results belongs.
            batch_size: Batch size for querying.

        Return:
            A dictionnary mapping each result name to its corresponding result summary.
        """
        results = {}
        for result_names_batch in batched(result_names, batch_size):
            request = CreateResultsMetaDataRequest(
                results=[
                    CreateResultsMetaDataRequest.ResultCreate(name=result_name)
                    for result_name in result_names_batch
                ],
                session_id=session_id,
            )
            response: CreateResultsMetaDataResponse = self._client.CreateResultsMetaData(request)
            for result_message in response.results:
                results[result_message.name] = Result.from_message(result_message)
        return results

    def create_results(
        self, results_data: Dict[str, bytes], session_id: str, batch_size: int = 1
    ) -> Dict[str, Result]:
        """Create multiple results from a mapping of result names and its corresponding data.

        Args:
            results_data: A dictionnary mapping the result names to their actual data.
            session_id: The ID of the session to create the results in.
            batch_size: Batch size for querying.

        Return:
            A dictionnary mapping each result name to its corresponding result summary.
        """
        results = {}
        for results_names_batch in batched(results_data.keys(), batch_size):
            request = CreateResultsRequest(
                results=[
                    CreateResultsRequest.ResultCreate(name=name, data=results_data[name])
                    for name in results_names_batch
                ],
                session_id=session_id,
            )
            response: CreateResultsResponse = self._client.CreateResults(request)
            for message in response.results:
                results[message.name] = Result.from_message(message)
        return results

    def upload_result_data(
        self, result_id: str, session_id: str, result_data: Union[bytes, bytearray]
    ) -> None:
        """Upload data for an empty result already created.

        Args:
            result_id: The ID of the result.
            result_data: The result data.
            session_id: The ID of the session.
        """
        data_chunk_max_size = self.get_service_config()

        def upload_result_stream():
            request = UploadResultDataRequest(
                id=UploadResultDataRequest.ResultIdentifier(
                    session_id=session_id, result_id=result_id
                )
            )
            yield request

            start = 0
            data_len = len(result_data)
            while start < data_len:
                chunk_size = min(data_chunk_max_size, data_len - start)
                request = UploadResultDataRequest(
                    data_chunk=result_data[start : start + chunk_size]
                )
                yield request
                start += chunk_size

        self._client.UploadResultData(upload_result_stream())

    def download_result_data(self, result_id: str, session_id: str) -> bytes:
        """Retrieve data of a result.

        Args:
            result_id: The ID of the result.
            session_id: The session of the result.

        Return:
            Result data.
        """
        request = DownloadResultDataRequest(result_id=result_id, session_id=session_id)
        streaming_call = self._client.DownloadResultData(request)
        return b"".join([message.data_chunk for message in streaming_call])

    def delete_result_data(
        self, result_ids: List[str], session_id: str, batch_size: int = 100
    ) -> None:
        """Delete data from multiple results

        Args:
            result_ids: The IDs of the results which data must be deleted.
            session_id: The ID of the session to which the results belongs.
            batch_size: Batch size for querying.
        """
        for result_ids_batch in batched(result_ids, batch_size):
            request = DeleteResultsDataRequest(result_id=result_ids_batch, session_id=session_id)
            self._client.DeleteResultsData(request)

    def get_service_config(self) -> int:
        """Get the configuration of the service.

        Return:
            Maximum size supported by a data chunk for the result service.
        """
        response: ResultsServiceConfigurationResponse = self._client.GetServiceConfiguration(
            Empty()
        )
        return response.data_chunk_max_size

    def watch_results(self):
        raise NotImplementedError()

    def import_data(
        self, session_id: str, results: Iterable[Tuple[str, bytes]]
    ) -> Dict[str, Result]:
        """
        Import existing data from the object storage into existing results.

        Args:
            session_id: The ID of the session.
            results: An iterable of tuples, where each tuple contains a result ID and an opaque ID.

        Returns:
            A dictionary mapping result IDs to Result objects.
        """
        result_opaque_ids = [
            ImportResultsDataRequest.ResultOpaqueId(result_id=result_id, opaque_id=opaque_id)
            for result_id, opaque_id in results
        ]
        response: ImportResultsDataResponse = self._client.ImportResultsData(
            ImportResultsDataRequest(
                session_id=session_id,
                results=result_opaque_ids,
            )
        )
        return {result.result_id: Result.from_message(result) for result in response.results}



from .partitions import ArmoniKPartitions, PartitionFieldFilter
from .sessions import ArmoniKSessions, SessionFieldFilter
from .tasks import ArmoniKTasks, TaskFieldFilter
from .results import ArmoniKResults, ResultFieldFilter
from .versions import ArmoniKVersions
from .events import ArmoniKEvents
from .health_checks import ArmoniKHealthChecks

__all__ = [
    "ArmoniKPartitions",
    "ArmoniKSessions",
    "ArmoniKTasks",
    "ArmoniKResults",
    "ArmoniKVersions",
    "ArmoniKEvents",
    "ArmoniKHealthChecks",
    "TaskFieldFilter",
    "PartitionFieldFilter",
    "SessionFieldFilter",
    "ResultFieldFilter",
]



#!/usr/bin/env python3
import itertools
from typing import Dict, cast, Union

import pytest

from dataclasses import dataclass

from armonik.common import Task, Partition
from armonik.common.filter import (
    Filter,
    StringFilter,
    BooleanFilter,
    NumberFilter,
    DisjunctionType,
    ConjunctionType,
    BasicMessageType,
    TaskOptionFilter,
    FilterError,
)
from armonik.protogen.common.filters_common_pb2 import FilterBoolean
from google.protobuf.message import Message


@dataclass
class DummyMessage(DisjunctionType):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._disjunction = []

    def __getattr__(self, item):
        if item == "or":
            return self._disjunction
        raise AttributeError()


@dataclass
class DummyMessageAnd(ConjunctionType):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._conjunction = []

    def __getattr__(self, item):
        if item == "and":
            return self._conjunction
        raise AttributeError()


@dataclass
class Field(Message):
    name: str = ""


@dataclass
class BaseMessage(BasicMessageType):
    def __init__(self, *, field, **kwargs):
        super().__init__(field=field, **kwargs)


@pytest.mark.parametrize(
    "filt,inverted",
    [
        (
            StringFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage) == "Test",
            StringFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage) != "Test",
        ),
        (
            StringFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage).contains("Test"),
            ~(StringFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage).contains("Test")),
        ),
        (
            BooleanFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage),
            BooleanFilter(Field(), DummyMessage, DummyMessageAnd, DummyMessage, value=False),
        ),
        (
            NumberFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage) > 0,
            NumberFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage) <= 0,
        ),
        (
            NumberFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage) >= 0,
            NumberFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage) < 0,
        ),
        (
            NumberFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage) < 0,
            NumberFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage) >= 0,
        ),
        (
            NumberFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage) <= 0,
            NumberFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage) > 0,
        ),
    ],
)
def test_inversion(filt: Filter, inverted: Filter):
    assert filt.operator != inverted.operator or filt.value == (
        not inverted.value
    )  # In case of BooleanFilter, the value is inverted, not the operator
    assert (~filt).operator == inverted.operator and (~filt).value == inverted.value
    assert filt.operator == (~(~filt)).operator and filt.value == (~(~filt)).value


@pytest.mark.parametrize(
    "filt",
    [
        (StringFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage).startswith("Test")),
        (StringFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage).endswith("Test")),
        (StringFilter(Field(), DummyMessage, DummyMessageAnd, BaseMessage)),  # No op
    ],
)
def test_inversion_raises(filt: Filter):
    with pytest.raises(Exception):
        test = ~filt
        print(test)


def test_descriptor():
    task = Task()
    assert task.id is None
    task.id = "xxx"
    assert isinstance(task.id, str)
    assert isinstance(Task.id, Filter)
    assert isinstance(Task.id == "xxx", Filter)


def test_extra_operator():
    with pytest.raises(FilterError):
        try:
            print((Task.id == "x") == "y")
        except FilterError as e:
            print(e)
            raise

    with pytest.raises(FilterError):
        try:
            print(((Task.id == "x") & (Task.session_id == "y")) == "y")
        except FilterError as e:
            print(e)
            raise


def test_extra_bool_compare():
    def check_eq(f1: Filter, f2: Filter) -> bool:
        return f1.operator == f2.operator and f1.value == f2.value

    task_filt = cast(Filter, Task.id == "x")
    task_filt_not = cast(Filter, Task.id != "x")

    assert isinstance(task_filt, Filter)
    assert isinstance(task_filt_not, Filter)

    task_filt_eq_true = cast(Filter, task_filt == True)  # noqa: E712
    task_filt_eq_false = cast(Filter, task_filt == False)  # noqa: E712
    task_filt_ne_true = cast(Filter, task_filt != True)  # noqa: E712
    task_filt_ne_false = cast(Filter, task_filt != False)  # noqa: E712
    assert check_eq(task_filt_eq_true, task_filt)
    assert not check_eq(task_filt_eq_true, task_filt_not)
    assert not check_eq(task_filt_eq_false, task_filt)
    assert check_eq(task_filt_eq_false, task_filt_not)
    assert not check_eq(task_filt_ne_true, task_filt)
    assert check_eq(task_filt_ne_true, task_filt_not)
    assert check_eq(task_filt_ne_false, task_filt)
    assert not check_eq(task_filt_ne_false, task_filt_not)


def test_task_option_descriptor():
    assert isinstance(Task.options, TaskOptionFilter)
    assert isinstance(Task.options.priority, Filter)
    with pytest.raises(FilterError):
        try:
            print(Task.options == "test")
        except FilterError as e:
            print(e)
            raise
    with pytest.raises(FilterError):
        try:
            print(Task.options.contains("test"))
        except FilterError as e:
            print(e)
            raise


def test_unavailable_field():
    with pytest.raises(FilterError):
        print(Partition.pod_configuration == {})


def apply_filter(state: Dict[str, bool], message: Union[Message, Filter]) -> bool:
    if isinstance(message, Filter):
        return apply_filter(state, message.to_message())
    if hasattr(message, "or"):
        return any(apply_filter(state, m) for m in getattr(message, "or"))
    if hasattr(message, "and"):
        return all(apply_filter(state, m) for m in getattr(message, "and"))
    vp = cast(FilterBoolean, getattr(message, "filter_boolean"))
    field = cast(Field, getattr(message, "field"))
    return vp.value == state[field.name]


def get_bool_filter(name: str) -> BooleanFilter:
    return BooleanFilter(Field(name), DummyMessage, DummyMessageAnd, BaseMessage)


def test_combine():
    a = get_bool_filter("a")
    b = get_bool_filter("b")
    c = get_bool_filter("c")
    d = get_bool_filter("d")
    e = get_bool_filter("e")
    for _a, _b, _c, _d, _e in itertools.product([True, False], repeat=5):
        t = {"a": _a, "b": _b, "c": _c, "d": _d, "e": _e}
        assert apply_filter(t, a) == _a
        assert apply_filter(t, (~a)) == (not _a)
        assert apply_filter(t, (a | b)) == (_a or _b)
        assert apply_filter(t, ((~a) | b)) == ((not _a) or _b)
        assert apply_filter(t, (a | (~b))) == (_a or not _b)
        assert apply_filter(t, ((~a) | (~b))) == (not _a or not _b)
        assert apply_filter(t, ((~a) & b)) == ((not _a) and _b)
        assert apply_filter(t, (~(a | b))) == (not (_a or _b))
        assert apply_filter(t, (~(a & b))) == (not (_a and _b))
        assert apply_filter(t, (c & (a | b))) == (_c and (_a or _b))
        assert apply_filter(t, (c | (a & b))) == (_c or (_a and _b))
        assert apply_filter(t, (a & (b | c)) & (d | e)) == (_a and (_b or _c) and (_d or _e))
        assert apply_filter(t, ~((a & (b | c)) & (d | e))) == (
            not (_a and (_b or _c) and (_d or _e))
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: applications_fields.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x61pplications_fields.proto\x12 armonik.api.grpc.v1.applications\"_\n\x13\x41pplicationRawField\x12H\n\x05\x66ield\x18\x01 \x01(\x0e\x32\x39.armonik.api.grpc.v1.applications.ApplicationRawEnumField\"o\n\x10\x41pplicationField\x12R\n\x11\x61pplication_field\x18\x01 \x01(\x0b\x32\x35.armonik.api.grpc.v1.applications.ApplicationRawFieldH\x00\x42\x07\n\x05\x66ield*\xe4\x01\n\x17\x41pplicationRawEnumField\x12*\n&APPLICATION_RAW_ENUM_FIELD_UNSPECIFIED\x10\x00\x12#\n\x1f\x41PPLICATION_RAW_ENUM_FIELD_NAME\x10\x01\x12&\n\"APPLICATION_RAW_ENUM_FIELD_VERSION\x10\x02\x12(\n$APPLICATION_RAW_ENUM_FIELD_NAMESPACE\x10\x03\x12&\n\"APPLICATION_RAW_ENUM_FIELD_SERVICE\x10\x04\x42M\n(fr.aneo.armonik.api.grpc.v1.applications\xaa\x02 ArmoniK.Api.gRPC.V1.Applicationsb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'applications_fields_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n(fr.aneo.armonik.api.grpc.v1.applications\252\002 ArmoniK.Api.gRPC.V1.Applications'
  _globals['_APPLICATIONRAWENUMFIELD']._serialized_start=274
  _globals['_APPLICATIONRAWENUMFIELD']._serialized_end=502
  _globals['_APPLICATIONRAWFIELD']._serialized_start=63
  _globals['_APPLICATIONRAWFIELD']._serialized_end=158
  _globals['_APPLICATIONFIELD']._serialized_start=160
  _globals['_APPLICATIONFIELD']._serialized_end=271
# @@protoc_insertion_point(module_scope)



from ._message_types import (
    DisjunctionType,
    ConjunctionType,
    BasicMessageType,
    InnerMessageType,
)
from .filter import (
    Filter,
    FilterError,
    NumberFilter,
    DurationFilter,
    DateFilter,
    StringFilter,
    ArrayFilter,
    BooleanFilter,
    StatusFilter,
    FilterDescriptor,
)

from ._filter_field import (
    PartitionFilter,
    TaskFilter,
    TaskOptionFilter,
    SessionFilter,
    SessionTaskOptionFilter,
    ResultFilter,
    ApplicationFilter,
    GenericTaskOptionsFilter,
    OutputFilter,
)

__all__ = [
    "Filter",
    "NumberFilter",
    "DurationFilter",
    "DateFilter",
    "StringFilter",
    "ArrayFilter",
    "BooleanFilter",
    "StatusFilter",
    "TaskFilter",
    "FilterError",
    "TaskOptionFilter",
    "DisjunctionType",
    "ConjunctionType",
    "BasicMessageType",
    "InnerMessageType",
    "SessionFilter",
    "SessionTaskOptionFilter",
    "ResultFilter",
    "PartitionFilter",
    "ApplicationFilter",
    "FilterDescriptor",
    "GenericTaskOptionsFilter",
    "OutputFilter",
]



from __future__ import annotations

import concurrent.futures
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Callable, cast, Iterable, List, Optional, Union, Collection

from grpc import Channel, RpcError

from ..common import (
    EventTypes,
    NewTaskEvent,
    NewResultEvent,
    ResultOwnerUpdateEvent,
    ResultStatusUpdateEvent,
    TaskStatusUpdateEvent,
    ResultStatus,
    Event,
    Result,
    batched,
)
from ..common.filter import Filter
from ..protogen.client.events_service_pb2_grpc import EventsStub
from ..protogen.common.events_common_pb2 import EventSubscriptionRequest, EventsEnum
from ..protogen.common.results_filters_pb2 import Filters as rawResultFilters
from ..protogen.common.tasks_filters_pb2 import Filters as rawTaskFilters


class ArmoniKEvents:
    _events_obj_mapping = {
        "new_result": NewResultEvent,
        "new_task": NewTaskEvent,
        "result_owner_update": ResultOwnerUpdateEvent,
        "result_status_update": ResultStatusUpdateEvent,
        "task_status_update": TaskStatusUpdateEvent,
    }

    def __init__(self, grpc_channel: Channel):
        """Events service client

        Args:
            grpc_channel: gRPC channel to use
        """
        self._client = EventsStub(grpc_channel)

    def get_events(
        self,
        session_id: str,
        event_types: Iterable[
            EventsEnum
        ],  # TODO: make EventTypes an enum when Python 3.8 support will be not supported
        event_handlers: List[Callable[[str, EventTypes, Event], bool]],
        task_filter: Optional[Filter] = None,
        result_filter: Optional[Filter] = None,
    ) -> None:
        """Get events that represents updates of result and tasks data.

        Args:
            session_id: The ID of the session.
            event_types: The list of the types of event to catch.
            event_handlers: The list of handlers that process the events. Handlers are evaluated in the order they are provided.
                An handler takes three positional arguments: the ID of the session, the type of event and the event as an object.
                An handler returns a boolean, if False the process continues, otherwise the stream is closed and the service stops
                listening to new events. As handlers are evaluated in order, when a handler interrupts execution (by returning True)
                all the handlers following it will not be executed.
            task_filter: A filter on tasks.
            result_filter: A filter on results.

        """
        request = EventSubscriptionRequest(
            session_id=session_id,
            returned_events=event_types,
            tasks_filters=cast(rawTaskFilters, task_filter.to_disjunction().to_message())
            if task_filter is not None
            else rawTaskFilters(),
            results_filters=cast(rawResultFilters, result_filter.to_disjunction().to_message())
            if result_filter is not None
            else rawResultFilters(),
        )

        streaming_call = self._client.GetEvents(request)
        for message in streaming_call:
            event_type = message.WhichOneof("update")
            if any(
                event_handler(
                    session_id,
                    EventTypes.from_string(event_type),
                    self._events_obj_mapping[event_type].from_raw_event(
                        getattr(message, event_type)
                    ),
                )
                for event_handler in event_handlers
            ):
                break

    def wait_for_result_availability(
        self,
        result_ids: Union[str, List[str]],
        session_id: str,
        bucket_size: int = 100,
        parallelism: int = 1,
    ) -> None:
        """Wait until a result is ready i.e its status updates to COMPLETED.

        Args:
            result_ids: The IDs of the results.
            session_id: The ID of the session.
            bucket_size: Batch size
            parallelism: Parallelism
        Raises:
            RuntimeError: If the result status is ABORTED.
        """
        if isinstance(result_ids, str):
            result_ids = [result_ids]
        result_ids = set(result_ids)
        if len(result_ids) == 0:
            return

        if parallelism > 1:
            pool = ThreadPoolExecutor(max_workers=parallelism)
            try:
                futures = [
                    pool.submit(_wait_all, self, session_id, batch)
                    for batch in batched(result_ids, bucket_size)
                ]
                for i, future in enumerate(concurrent.futures.as_completed(futures)):
                    exp = future.exception()
                    if exp is not None:
                        for f in futures:
                            f.cancel()
                        raise exp
            finally:
                pool.shutdown(wait=False)
        else:
            for batch in batched(result_ids, bucket_size):
                _wait_all(self, session_id, batch)


def _wait_all(event_client: ArmoniKEvents, session_id: str, results: Collection[str]):
    if len(results) == 0:
        return
    results_filter = None
    for result_id in results:
        results_filter = (
            Result.result_id == result_id
            if results_filter is None
            else (results_filter | (Result.result_id == result_id))
        )

    not_found = set(results)

    def handler(_, _2, event: Event) -> bool:
        event = cast(Union[NewResultEvent, ResultStatusUpdateEvent], event)
        if event.result_id in not_found:
            if event.status == ResultStatus.COMPLETED:
                not_found.remove(event.result_id)
                if not not_found:
                    return True
            elif event.status == ResultStatus.ABORTED:
                raise RuntimeError(f"Result {event.result_id} has been aborted.")
        return False

    while not_found:
        try:
            event_client.get_events(
                session_id,
                [EventTypes.RESULT_STATUS_UPDATE, EventTypes.NEW_RESULT],
                [handler],
                None,
                results_filter,
            )
        except RpcError:
            pass
        else:
            break



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor

class SessionStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SESSION_STATUS_UNSPECIFIED: _ClassVar[SessionStatus]
    SESSION_STATUS_RUNNING: _ClassVar[SessionStatus]
    SESSION_STATUS_CANCELLED: _ClassVar[SessionStatus]
    SESSION_STATUS_PAUSED: _ClassVar[SessionStatus]
    SESSION_STATUS_CLOSED: _ClassVar[SessionStatus]
    SESSION_STATUS_PURGED: _ClassVar[SessionStatus]
    SESSION_STATUS_DELETED: _ClassVar[SessionStatus]
SESSION_STATUS_UNSPECIFIED: SessionStatus
SESSION_STATUS_RUNNING: SessionStatus
SESSION_STATUS_CANCELLED: SessionStatus
SESSION_STATUS_PAUSED: SessionStatus
SESSION_STATUS_CLOSED: SessionStatus
SESSION_STATUS_PURGED: SessionStatus
SESSION_STATUS_DELETED: SessionStatus



from .worker import ArmoniKWorker, armonik_worker
from .taskhandler import TaskHandler
from .seqlogger import ClefLogger

__all__ = [
    "ArmoniKWorker",
    "TaskHandler",
    "ClefLogger",
    "armonik_worker",
]



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SessionRawEnumField(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SESSION_RAW_ENUM_FIELD_UNSPECIFIED: _ClassVar[SessionRawEnumField]
    SESSION_RAW_ENUM_FIELD_SESSION_ID: _ClassVar[SessionRawEnumField]
    SESSION_RAW_ENUM_FIELD_STATUS: _ClassVar[SessionRawEnumField]
    SESSION_RAW_ENUM_FIELD_PARTITION_IDS: _ClassVar[SessionRawEnumField]
    SESSION_RAW_ENUM_FIELD_OPTIONS: _ClassVar[SessionRawEnumField]
    SESSION_RAW_ENUM_FIELD_CREATED_AT: _ClassVar[SessionRawEnumField]
    SESSION_RAW_ENUM_FIELD_CANCELLED_AT: _ClassVar[SessionRawEnumField]
    SESSION_RAW_ENUM_FIELD_CLOSED_AT: _ClassVar[SessionRawEnumField]
    SESSION_RAW_ENUM_FIELD_PURGED_AT: _ClassVar[SessionRawEnumField]
    SESSION_RAW_ENUM_FIELD_DELETED_AT: _ClassVar[SessionRawEnumField]
    SESSION_RAW_ENUM_FIELD_DURATION: _ClassVar[SessionRawEnumField]
    SESSION_RAW_ENUM_FIELD_WORKER_SUBMISSION: _ClassVar[SessionRawEnumField]
    SESSION_RAW_ENUM_FIELD_CLIENT_SUBMISSION: _ClassVar[SessionRawEnumField]

class TaskOptionEnumField(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    TASK_OPTION_ENUM_FIELD_UNSPECIFIED: _ClassVar[TaskOptionEnumField]
    TASK_OPTION_ENUM_FIELD_MAX_DURATION: _ClassVar[TaskOptionEnumField]
    TASK_OPTION_ENUM_FIELD_MAX_RETRIES: _ClassVar[TaskOptionEnumField]
    TASK_OPTION_ENUM_FIELD_PRIORITY: _ClassVar[TaskOptionEnumField]
    TASK_OPTION_ENUM_FIELD_PARTITION_ID: _ClassVar[TaskOptionEnumField]
    TASK_OPTION_ENUM_FIELD_APPLICATION_NAME: _ClassVar[TaskOptionEnumField]
    TASK_OPTION_ENUM_FIELD_APPLICATION_VERSION: _ClassVar[TaskOptionEnumField]
    TASK_OPTION_ENUM_FIELD_APPLICATION_NAMESPACE: _ClassVar[TaskOptionEnumField]
    TASK_OPTION_ENUM_FIELD_APPLICATION_SERVICE: _ClassVar[TaskOptionEnumField]
    TASK_OPTION_ENUM_FIELD_ENGINE_TYPE: _ClassVar[TaskOptionEnumField]
SESSION_RAW_ENUM_FIELD_UNSPECIFIED: SessionRawEnumField
SESSION_RAW_ENUM_FIELD_SESSION_ID: SessionRawEnumField
SESSION_RAW_ENUM_FIELD_STATUS: SessionRawEnumField
SESSION_RAW_ENUM_FIELD_PARTITION_IDS: SessionRawEnumField
SESSION_RAW_ENUM_FIELD_OPTIONS: SessionRawEnumField
SESSION_RAW_ENUM_FIELD_CREATED_AT: SessionRawEnumField
SESSION_RAW_ENUM_FIELD_CANCELLED_AT: SessionRawEnumField
SESSION_RAW_ENUM_FIELD_CLOSED_AT: SessionRawEnumField
SESSION_RAW_ENUM_FIELD_PURGED_AT: SessionRawEnumField
SESSION_RAW_ENUM_FIELD_DELETED_AT: SessionRawEnumField
SESSION_RAW_ENUM_FIELD_DURATION: SessionRawEnumField
SESSION_RAW_ENUM_FIELD_WORKER_SUBMISSION: SessionRawEnumField
SESSION_RAW_ENUM_FIELD_CLIENT_SUBMISSION: SessionRawEnumField
TASK_OPTION_ENUM_FIELD_UNSPECIFIED: TaskOptionEnumField
TASK_OPTION_ENUM_FIELD_MAX_DURATION: TaskOptionEnumField
TASK_OPTION_ENUM_FIELD_MAX_RETRIES: TaskOptionEnumField
TASK_OPTION_ENUM_FIELD_PRIORITY: TaskOptionEnumField
TASK_OPTION_ENUM_FIELD_PARTITION_ID: TaskOptionEnumField
TASK_OPTION_ENUM_FIELD_APPLICATION_NAME: TaskOptionEnumField
TASK_OPTION_ENUM_FIELD_APPLICATION_VERSION: TaskOptionEnumField
TASK_OPTION_ENUM_FIELD_APPLICATION_NAMESPACE: TaskOptionEnumField
TASK_OPTION_ENUM_FIELD_APPLICATION_SERVICE: TaskOptionEnumField
TASK_OPTION_ENUM_FIELD_ENGINE_TYPE: TaskOptionEnumField

class SessionRawField(_message.Message):
    __slots__ = ("field",)
    FIELD_FIELD_NUMBER: _ClassVar[int]
    field: SessionRawEnumField
    def __init__(self, field: _Optional[_Union[SessionRawEnumField, str]] = ...) -> None: ...

class TaskOptionField(_message.Message):
    __slots__ = ("field",)
    FIELD_FIELD_NUMBER: _ClassVar[int]
    field: TaskOptionEnumField
    def __init__(self, field: _Optional[_Union[TaskOptionEnumField, str]] = ...) -> None: ...

class TaskOptionGenericField(_message.Message):
    __slots__ = ("field",)
    FIELD_FIELD_NUMBER: _ClassVar[int]
    field: str
    def __init__(self, field: _Optional[str] = ...) -> None: ...

class SessionField(_message.Message):
    __slots__ = ("session_raw_field", "task_option_field", "task_option_generic_field")
    SESSION_RAW_FIELD_FIELD_NUMBER: _ClassVar[int]
    TASK_OPTION_FIELD_FIELD_NUMBER: _ClassVar[int]
    TASK_OPTION_GENERIC_FIELD_FIELD_NUMBER: _ClassVar[int]
    session_raw_field: SessionRawField
    task_option_field: TaskOptionField
    task_option_generic_field: TaskOptionGenericField
    def __init__(self, session_raw_field: _Optional[_Union[SessionRawField, _Mapping]] = ..., task_option_field: _Optional[_Union[TaskOptionField, _Mapping]] = ..., task_option_generic_field: _Optional[_Union[TaskOptionGenericField, _Mapping]] = ...) -> None: ...



from __future__ import annotations

from typing import List, Optional, Tuple, cast

from deprecation import deprecated
from grpc import Channel

from .. import __version__
from ..common import Direction, Session, TaskOptions
from ..common.filter import Filter, StringFilter, SessionFilter
from ..protogen.client.sessions_service_pb2_grpc import SessionsStub
from ..protogen.common.sessions_common_pb2 import (
    CancelSessionRequest,
    CancelSessionResponse,
    CreateSessionRequest,
    DeleteSessionRequest,
    DeleteSessionResponse,
    GetSessionRequest,
    GetSessionResponse,
    PauseSessionRequest,
    PauseSessionResponse,
    PurgeSessionRequest,
    PurgeSessionResponse,
    CloseSessionRequest,
    CloseSessionResponse,
    ResumeSessionRequest,
    ResumeSessionResponse,
    StopSubmissionRequest,
    StopSubmissionResponse,
    ListSessionsRequest,
    ListSessionsResponse,
)
from ..protogen.common.sessions_fields_pb2 import (
    SessionField,
)
from ..protogen.common.sort_direction_pb2 import SortDirection


@deprecated("3.19.0", None, __version__, "Use Session.<name of the field> instead")
class SessionFieldFilter:
    """
    Enumeration of the available filters
    """

    STATUS = Session.status

    @staticmethod
    def task_options_key(option_key: str) -> StringFilter:
        """
        Filter for the TaskOptions.Options dictionary
        Args:
            option_key: key in the dictionary

        Returns:
            Corresponding filter
        """
        return Session.options[option_key]


class ArmoniKSessions:
    def __init__(self, grpc_channel: Channel):
        """Session service client

        Args:
            grpc_channel: gRPC channel to use
        """
        self._client = SessionsStub(grpc_channel)

    def create_session(
        self,
        default_task_options: TaskOptions,
        partition_ids: Optional[List[str]] = None,
    ) -> str:
        """Create a session

        Args:
            default_task_options: Default TaskOptions used when
                submitting tasks without specifying the options
            partition_ids: List of partitions this session can send
                tasks to. If unspecified, can only send to the default
                partition

        Returns:
            Session Id
        """
        request = CreateSessionRequest(
            default_task_option=default_task_options.to_message(),
            partition_ids=partition_ids if partition_ids else [],
        )
        return self._client.CreateSession(request).session_id

    def get_session(self, session_id: str):
        """Get a session by its ID.

        Args:
            session_id: The ID of the session.

        Return:
            The session summary.
        """
        request = GetSessionRequest(session_id=session_id)
        response: GetSessionResponse = self._client.GetSession(request)
        return Session.from_message(response.session)

    def list_sessions(
        self,
        session_filter: Optional[Filter] = None,
        page: int = 0,
        page_size: int = 1000,
        sort_field: Filter = Session.status,
        sort_direction: SortDirection = Direction.ASC,
    ) -> Tuple[int, List[Session]]:
        """
        List sessions

        Args:
            session_filter (Filter): Filter to apply when listing sessions.
            page: page number to request, this can be useful when paginating the result, defaults to 0
            page_size: size of a page, defaults to 1000
            sort_field: field on which to sort the resulting list, defaults to the status
            sort_direction: direction of the sort, defaults to ascending

        Returns:
            A tuple containing :
            - The total number of sessions for the given filter
            - The obtained list of sessions
        """
        request = ListSessionsRequest(
            page=page,
            page_size=page_size,
            filters=(
                SessionFilter().to_message()
                if session_filter is None
                else session_filter.to_message()
            ),
            sort=ListSessionsRequest.Sort(
                field=cast(SessionField, sort_field.field), direction=sort_direction
            ),
        )
        response: ListSessionsResponse = self._client.ListSessions(request)
        return response.total, [Session.from_message(s) for s in response.sessions]

    def cancel_session(self, session_id: str) -> Session:
        """Cancel a session

        Args:
            session_id: Id of the session to be cancelled
        """
        request = CancelSessionRequest(session_id=session_id)
        response: CancelSessionResponse = self._client.CancelSession(request)
        return Session.from_message(response.session)

    def pause_session(self, session_id: str) -> Session:
        """Pause a session by its id.

        Args:
            session_id: Id of the session to be paused.

        Returns:
            session metadata
        """
        request = PauseSessionRequest(session_id=session_id)
        response: PauseSessionResponse = self._client.PauseSession(request)
        return Session.from_message(response.session)

    def resume_session(self, session_id: str) -> Session:
        """Resume a session by its id.

        Args:
            session_id: Id of the session to be resumed.

        Returns:
            session metadata
        """
        request = ResumeSessionRequest(session_id=session_id)
        response: ResumeSessionResponse = self._client.ResumeSession(request)
        return Session.from_message(response.session)

    def close_session(self, session_id: str) -> Session:
        """Close a session by its id.

        Args:
            session_id: Id of the session to be closed.

        Returns:
            session metadata
        """
        request = CloseSessionRequest(session_id=session_id)
        response: CloseSessionResponse = self._client.CloseSession(request)
        return Session.from_message(response.session)

    def purge_session(self, session_id: str) -> Session:
        """Purge a session by its id.

        Args:
            session_id: Id of the session to be purged.

        Returns:
            session metadata
        """
        request = PurgeSessionRequest(session_id=session_id)
        response: PurgeSessionResponse = self._client.PurgeSession(request)
        return Session.from_message(response.session)

    def delete_session(self, session_id: str) -> Session:
        """Delete a session by its id.

        Args:
            session_id: Id of the session to be deleted.

        Returns:
            session metadata
        """
        request = DeleteSessionRequest(session_id=session_id)
        response: DeleteSessionResponse = self._client.DeleteSession(request)
        return Session.from_message(response.session)

    def stop_submission_session(self, session_id: str, client: bool, worker: bool) -> Session:
        """Stops clients and/or workers from submitting new tasks in the given session.

        Args:
            session_id: Id of the session.
            client: Stops clients from submitting new tasks in the given session.
            worker: Stops workers from submitting new tasks in the given session.

        Returns:
            session metadata
        """
        request = StopSubmissionRequest(session_id=session_id, client=client, worker=worker)
        response: StopSubmissionResponse = self._client.StopSubmission(request)
        return Session.from_message(response.session)



from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from . import objects_pb2 as _objects_pb2
from . import sort_direction_pb2 as _sort_direction_pb2
from . import task_status_pb2 as _task_status_pb2
from . import tasks_fields_pb2 as _tasks_fields_pb2
from . import tasks_filters_pb2 as _tasks_filters_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class TaskDetailed(_message.Message):
    __slots__ = ("id", "session_id", "owner_pod_id", "initial_task_id", "parent_task_ids", "data_dependencies", "expected_output_ids", "retry_of_ids", "status", "status_message", "options", "created_at", "submitted_at", "received_at", "acquired_at", "fetched_at", "started_at", "processed_at", "ended_at", "pod_ttl", "creation_to_end_duration", "processing_to_end_duration", "received_to_end_duration", "payload_id", "created_by", "output", "pod_hostname")
    class Output(_message.Message):
        __slots__ = ("success", "error")
        SUCCESS_FIELD_NUMBER: _ClassVar[int]
        ERROR_FIELD_NUMBER: _ClassVar[int]
        success: bool
        error: str
        def __init__(self, success: bool = ..., error: _Optional[str] = ...) -> None: ...
    ID_FIELD_NUMBER: _ClassVar[int]
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    OWNER_POD_ID_FIELD_NUMBER: _ClassVar[int]
    INITIAL_TASK_ID_FIELD_NUMBER: _ClassVar[int]
    PARENT_TASK_IDS_FIELD_NUMBER: _ClassVar[int]
    DATA_DEPENDENCIES_FIELD_NUMBER: _ClassVar[int]
    EXPECTED_OUTPUT_IDS_FIELD_NUMBER: _ClassVar[int]
    RETRY_OF_IDS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    STATUS_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    SUBMITTED_AT_FIELD_NUMBER: _ClassVar[int]
    RECEIVED_AT_FIELD_NUMBER: _ClassVar[int]
    ACQUIRED_AT_FIELD_NUMBER: _ClassVar[int]
    FETCHED_AT_FIELD_NUMBER: _ClassVar[int]
    STARTED_AT_FIELD_NUMBER: _ClassVar[int]
    PROCESSED_AT_FIELD_NUMBER: _ClassVar[int]
    ENDED_AT_FIELD_NUMBER: _ClassVar[int]
    POD_TTL_FIELD_NUMBER: _ClassVar[int]
    CREATION_TO_END_DURATION_FIELD_NUMBER: _ClassVar[int]
    PROCESSING_TO_END_DURATION_FIELD_NUMBER: _ClassVar[int]
    RECEIVED_TO_END_DURATION_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_ID_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_FIELD_NUMBER: _ClassVar[int]
    POD_HOSTNAME_FIELD_NUMBER: _ClassVar[int]
    id: str
    session_id: str
    owner_pod_id: str
    initial_task_id: str
    parent_task_ids: _containers.RepeatedScalarFieldContainer[str]
    data_dependencies: _containers.RepeatedScalarFieldContainer[str]
    expected_output_ids: _containers.RepeatedScalarFieldContainer[str]
    retry_of_ids: _containers.RepeatedScalarFieldContainer[str]
    status: _task_status_pb2.TaskStatus
    status_message: str
    options: _objects_pb2.TaskOptions
    created_at: _timestamp_pb2.Timestamp
    submitted_at: _timestamp_pb2.Timestamp
    received_at: _timestamp_pb2.Timestamp
    acquired_at: _timestamp_pb2.Timestamp
    fetched_at: _timestamp_pb2.Timestamp
    started_at: _timestamp_pb2.Timestamp
    processed_at: _timestamp_pb2.Timestamp
    ended_at: _timestamp_pb2.Timestamp
    pod_ttl: _timestamp_pb2.Timestamp
    creation_to_end_duration: _duration_pb2.Duration
    processing_to_end_duration: _duration_pb2.Duration
    received_to_end_duration: _duration_pb2.Duration
    payload_id: str
    created_by: str
    output: TaskDetailed.Output
    pod_hostname: str
    def __init__(self, id: _Optional[str] = ..., session_id: _Optional[str] = ..., owner_pod_id: _Optional[str] = ..., initial_task_id: _Optional[str] = ..., parent_task_ids: _Optional[_Iterable[str]] = ..., data_dependencies: _Optional[_Iterable[str]] = ..., expected_output_ids: _Optional[_Iterable[str]] = ..., retry_of_ids: _Optional[_Iterable[str]] = ..., status: _Optional[_Union[_task_status_pb2.TaskStatus, str]] = ..., status_message: _Optional[str] = ..., options: _Optional[_Union[_objects_pb2.TaskOptions, _Mapping]] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., submitted_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., received_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., acquired_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., fetched_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., started_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., processed_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., ended_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., pod_ttl: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., creation_to_end_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., processing_to_end_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., received_to_end_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., payload_id: _Optional[str] = ..., created_by: _Optional[str] = ..., output: _Optional[_Union[TaskDetailed.Output, _Mapping]] = ..., pod_hostname: _Optional[str] = ...) -> None: ...

class TaskSummary(_message.Message):
    __slots__ = ("id", "session_id", "owner_pod_id", "initial_task_id", "count_parent_task_ids", "count_data_dependencies", "count_expected_output_ids", "count_retry_of_ids", "status", "status_message", "options", "created_at", "submitted_at", "received_at", "acquired_at", "fetched_at", "started_at", "processed_at", "ended_at", "pod_ttl", "creation_to_end_duration", "processing_to_end_duration", "received_to_end_duration", "payload_id", "created_by", "error", "pod_hostname")
    ID_FIELD_NUMBER: _ClassVar[int]
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    OWNER_POD_ID_FIELD_NUMBER: _ClassVar[int]
    INITIAL_TASK_ID_FIELD_NUMBER: _ClassVar[int]
    COUNT_PARENT_TASK_IDS_FIELD_NUMBER: _ClassVar[int]
    COUNT_DATA_DEPENDENCIES_FIELD_NUMBER: _ClassVar[int]
    COUNT_EXPECTED_OUTPUT_IDS_FIELD_NUMBER: _ClassVar[int]
    COUNT_RETRY_OF_IDS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    STATUS_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    SUBMITTED_AT_FIELD_NUMBER: _ClassVar[int]
    RECEIVED_AT_FIELD_NUMBER: _ClassVar[int]
    ACQUIRED_AT_FIELD_NUMBER: _ClassVar[int]
    FETCHED_AT_FIELD_NUMBER: _ClassVar[int]
    STARTED_AT_FIELD_NUMBER: _ClassVar[int]
    PROCESSED_AT_FIELD_NUMBER: _ClassVar[int]
    ENDED_AT_FIELD_NUMBER: _ClassVar[int]
    POD_TTL_FIELD_NUMBER: _ClassVar[int]
    CREATION_TO_END_DURATION_FIELD_NUMBER: _ClassVar[int]
    PROCESSING_TO_END_DURATION_FIELD_NUMBER: _ClassVar[int]
    RECEIVED_TO_END_DURATION_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_ID_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    POD_HOSTNAME_FIELD_NUMBER: _ClassVar[int]
    id: str
    session_id: str
    owner_pod_id: str
    initial_task_id: str
    count_parent_task_ids: int
    count_data_dependencies: int
    count_expected_output_ids: int
    count_retry_of_ids: int
    status: _task_status_pb2.TaskStatus
    status_message: str
    options: _objects_pb2.TaskOptions
    created_at: _timestamp_pb2.Timestamp
    submitted_at: _timestamp_pb2.Timestamp
    received_at: _timestamp_pb2.Timestamp
    acquired_at: _timestamp_pb2.Timestamp
    fetched_at: _timestamp_pb2.Timestamp
    started_at: _timestamp_pb2.Timestamp
    processed_at: _timestamp_pb2.Timestamp
    ended_at: _timestamp_pb2.Timestamp
    pod_ttl: _timestamp_pb2.Timestamp
    creation_to_end_duration: _duration_pb2.Duration
    processing_to_end_duration: _duration_pb2.Duration
    received_to_end_duration: _duration_pb2.Duration
    payload_id: str
    created_by: str
    error: str
    pod_hostname: str
    def __init__(self, id: _Optional[str] = ..., session_id: _Optional[str] = ..., owner_pod_id: _Optional[str] = ..., initial_task_id: _Optional[str] = ..., count_parent_task_ids: _Optional[int] = ..., count_data_dependencies: _Optional[int] = ..., count_expected_output_ids: _Optional[int] = ..., count_retry_of_ids: _Optional[int] = ..., status: _Optional[_Union[_task_status_pb2.TaskStatus, str]] = ..., status_message: _Optional[str] = ..., options: _Optional[_Union[_objects_pb2.TaskOptions, _Mapping]] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., submitted_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., received_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., acquired_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., fetched_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., started_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., processed_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., ended_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., pod_ttl: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., creation_to_end_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., processing_to_end_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., received_to_end_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., payload_id: _Optional[str] = ..., created_by: _Optional[str] = ..., error: _Optional[str] = ..., pod_hostname: _Optional[str] = ...) -> None: ...

class ListTasksRequest(_message.Message):
    __slots__ = ("page", "page_size", "filters", "sort", "with_errors")
    class Sort(_message.Message):
        __slots__ = ("field", "direction")
        FIELD_FIELD_NUMBER: _ClassVar[int]
        DIRECTION_FIELD_NUMBER: _ClassVar[int]
        field: _tasks_fields_pb2.TaskField
        direction: _sort_direction_pb2.SortDirection
        def __init__(self, field: _Optional[_Union[_tasks_fields_pb2.TaskField, _Mapping]] = ..., direction: _Optional[_Union[_sort_direction_pb2.SortDirection, str]] = ...) -> None: ...
    PAGE_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    SORT_FIELD_NUMBER: _ClassVar[int]
    WITH_ERRORS_FIELD_NUMBER: _ClassVar[int]
    page: int
    page_size: int
    filters: _tasks_filters_pb2.Filters
    sort: ListTasksRequest.Sort
    with_errors: bool
    def __init__(self, page: _Optional[int] = ..., page_size: _Optional[int] = ..., filters: _Optional[_Union[_tasks_filters_pb2.Filters, _Mapping]] = ..., sort: _Optional[_Union[ListTasksRequest.Sort, _Mapping]] = ..., with_errors: bool = ...) -> None: ...

class ListTasksResponse(_message.Message):
    __slots__ = ("tasks", "page", "page_size", "total")
    TASKS_FIELD_NUMBER: _ClassVar[int]
    PAGE_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    TOTAL_FIELD_NUMBER: _ClassVar[int]
    tasks: _containers.RepeatedCompositeFieldContainer[TaskSummary]
    page: int
    page_size: int
    total: int
    def __init__(self, tasks: _Optional[_Iterable[_Union[TaskSummary, _Mapping]]] = ..., page: _Optional[int] = ..., page_size: _Optional[int] = ..., total: _Optional[int] = ...) -> None: ...

class ListTasksDetailedResponse(_message.Message):
    __slots__ = ("tasks", "page", "page_size", "total")
    TASKS_FIELD_NUMBER: _ClassVar[int]
    PAGE_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    TOTAL_FIELD_NUMBER: _ClassVar[int]
    tasks: _containers.RepeatedCompositeFieldContainer[TaskDetailed]
    page: int
    page_size: int
    total: int
    def __init__(self, tasks: _Optional[_Iterable[_Union[TaskDetailed, _Mapping]]] = ..., page: _Optional[int] = ..., page_size: _Optional[int] = ..., total: _Optional[int] = ...) -> None: ...

class GetTaskRequest(_message.Message):
    __slots__ = ("task_id",)
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    def __init__(self, task_id: _Optional[str] = ...) -> None: ...

class GetTaskResponse(_message.Message):
    __slots__ = ("task",)
    TASK_FIELD_NUMBER: _ClassVar[int]
    task: TaskDetailed
    def __init__(self, task: _Optional[_Union[TaskDetailed, _Mapping]] = ...) -> None: ...

class CancelTasksRequest(_message.Message):
    __slots__ = ("task_ids",)
    TASK_IDS_FIELD_NUMBER: _ClassVar[int]
    task_ids: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, task_ids: _Optional[_Iterable[str]] = ...) -> None: ...

class CancelTasksResponse(_message.Message):
    __slots__ = ("tasks",)
    TASKS_FIELD_NUMBER: _ClassVar[int]
    tasks: _containers.RepeatedCompositeFieldContainer[TaskSummary]
    def __init__(self, tasks: _Optional[_Iterable[_Union[TaskSummary, _Mapping]]] = ...) -> None: ...

class GetResultIdsRequest(_message.Message):
    __slots__ = ("task_id",)
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    task_id: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, task_id: _Optional[_Iterable[str]] = ...) -> None: ...

class GetResultIdsResponse(_message.Message):
    __slots__ = ("task_results",)
    class MapTaskResult(_message.Message):
        __slots__ = ("task_id", "result_ids")
        TASK_ID_FIELD_NUMBER: _ClassVar[int]
        RESULT_IDS_FIELD_NUMBER: _ClassVar[int]
        task_id: str
        result_ids: _containers.RepeatedScalarFieldContainer[str]
        def __init__(self, task_id: _Optional[str] = ..., result_ids: _Optional[_Iterable[str]] = ...) -> None: ...
    TASK_RESULTS_FIELD_NUMBER: _ClassVar[int]
    task_results: _containers.RepeatedCompositeFieldContainer[GetResultIdsResponse.MapTaskResult]
    def __init__(self, task_results: _Optional[_Iterable[_Union[GetResultIdsResponse.MapTaskResult, _Mapping]]] = ...) -> None: ...

class CountTasksByStatusRequest(_message.Message):
    __slots__ = ("filters",)
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    filters: _tasks_filters_pb2.Filters
    def __init__(self, filters: _Optional[_Union[_tasks_filters_pb2.Filters, _Mapping]] = ...) -> None: ...

class CountTasksByStatusResponse(_message.Message):
    __slots__ = ("status",)
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: _containers.RepeatedCompositeFieldContainer[_objects_pb2.StatusCount]
    def __init__(self, status: _Optional[_Iterable[_Union[_objects_pb2.StatusCount, _Mapping]]] = ...) -> None: ...

class SubmitTasksRequest(_message.Message):
    __slots__ = ("session_id", "task_options", "task_creations")
    class TaskCreation(_message.Message):
        __slots__ = ("expected_output_keys", "data_dependencies", "payload_id", "task_options")
        EXPECTED_OUTPUT_KEYS_FIELD_NUMBER: _ClassVar[int]
        DATA_DEPENDENCIES_FIELD_NUMBER: _ClassVar[int]
        PAYLOAD_ID_FIELD_NUMBER: _ClassVar[int]
        TASK_OPTIONS_FIELD_NUMBER: _ClassVar[int]
        expected_output_keys: _containers.RepeatedScalarFieldContainer[str]
        data_dependencies: _containers.RepeatedScalarFieldContainer[str]
        payload_id: str
        task_options: _objects_pb2.TaskOptions
        def __init__(self, expected_output_keys: _Optional[_Iterable[str]] = ..., data_dependencies: _Optional[_Iterable[str]] = ..., payload_id: _Optional[str] = ..., task_options: _Optional[_Union[_objects_pb2.TaskOptions, _Mapping]] = ...) -> None: ...
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    TASK_OPTIONS_FIELD_NUMBER: _ClassVar[int]
    TASK_CREATIONS_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    task_options: _objects_pb2.TaskOptions
    task_creations: _containers.RepeatedCompositeFieldContainer[SubmitTasksRequest.TaskCreation]
    def __init__(self, session_id: _Optional[str] = ..., task_options: _Optional[_Union[_objects_pb2.TaskOptions, _Mapping]] = ..., task_creations: _Optional[_Iterable[_Union[SubmitTasksRequest.TaskCreation, _Mapping]]] = ...) -> None: ...

class SubmitTasksResponse(_message.Message):
    __slots__ = ("task_infos",)
    class TaskInfo(_message.Message):
        __slots__ = ("task_id", "expected_output_ids", "data_dependencies", "payload_id")
        TASK_ID_FIELD_NUMBER: _ClassVar[int]
        EXPECTED_OUTPUT_IDS_FIELD_NUMBER: _ClassVar[int]
        DATA_DEPENDENCIES_FIELD_NUMBER: _ClassVar[int]
        PAYLOAD_ID_FIELD_NUMBER: _ClassVar[int]
        task_id: str
        expected_output_ids: _containers.RepeatedScalarFieldContainer[str]
        data_dependencies: _containers.RepeatedScalarFieldContainer[str]
        payload_id: str
        def __init__(self, task_id: _Optional[str] = ..., expected_output_ids: _Optional[_Iterable[str]] = ..., data_dependencies: _Optional[_Iterable[str]] = ..., payload_id: _Optional[str] = ...) -> None: ...
    TASK_INFOS_FIELD_NUMBER: _ClassVar[int]
    task_infos: _containers.RepeatedCompositeFieldContainer[SubmitTasksResponse.TaskInfo]
    def __init__(self, task_infos: _Optional[_Iterable[_Union[SubmitTasksResponse.TaskInfo, _Mapping]]] = ...) -> None: ...



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class PartitionRawEnumField(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    PARTITION_RAW_ENUM_FIELD_UNSPECIFIED: _ClassVar[PartitionRawEnumField]
    PARTITION_RAW_ENUM_FIELD_ID: _ClassVar[PartitionRawEnumField]
    PARTITION_RAW_ENUM_FIELD_PARENT_PARTITION_IDS: _ClassVar[PartitionRawEnumField]
    PARTITION_RAW_ENUM_FIELD_POD_RESERVED: _ClassVar[PartitionRawEnumField]
    PARTITION_RAW_ENUM_FIELD_POD_MAX: _ClassVar[PartitionRawEnumField]
    PARTITION_RAW_ENUM_FIELD_PREEMPTION_PERCENTAGE: _ClassVar[PartitionRawEnumField]
    PARTITION_RAW_ENUM_FIELD_PRIORITY: _ClassVar[PartitionRawEnumField]
PARTITION_RAW_ENUM_FIELD_UNSPECIFIED: PartitionRawEnumField
PARTITION_RAW_ENUM_FIELD_ID: PartitionRawEnumField
PARTITION_RAW_ENUM_FIELD_PARENT_PARTITION_IDS: PartitionRawEnumField
PARTITION_RAW_ENUM_FIELD_POD_RESERVED: PartitionRawEnumField
PARTITION_RAW_ENUM_FIELD_POD_MAX: PartitionRawEnumField
PARTITION_RAW_ENUM_FIELD_PREEMPTION_PERCENTAGE: PartitionRawEnumField
PARTITION_RAW_ENUM_FIELD_PRIORITY: PartitionRawEnumField

class PartitionRawField(_message.Message):
    __slots__ = ("field",)
    FIELD_FIELD_NUMBER: _ClassVar[int]
    field: PartitionRawEnumField
    def __init__(self, field: _Optional[_Union[PartitionRawEnumField, str]] = ...) -> None: ...

class PartitionField(_message.Message):
    __slots__ = ("partition_raw_field",)
    PARTITION_RAW_FIELD_FIELD_NUMBER: _ClassVar[int]
    partition_raw_field: PartitionRawField
    def __init__(self, partition_raw_field: _Optional[_Union[PartitionRawField, _Mapping]] = ...) -> None: ...



from __future__ import annotations

import os
from collections.abc import Mapping
from typing import Optional, Dict, List, Union, KeysView

from ..common import TaskOptions, TaskDefinition, Result, Task
from ..common.helpers import batched
from ..protogen.common.agent_common_pb2 import (
    CreateResultsMetaDataRequest,
    CreateResultsMetaDataResponse,
    NotifyResultDataRequest,
    CreateResultsRequest,
    CreateResultsResponse,
    SubmitTasksRequest,
)
from ..protogen.common.objects_pb2 import (
    Configuration,
)
from ..protogen.common.worker_common_pb2 import ProcessRequest
from ..protogen.worker.agent_service_pb2_grpc import AgentStub


class LazyLoadDict(Mapping):
    def __init__(self, data_folder: str, ids: List[str]):
        self.__data_folder = data_folder
        self._data: Dict[str, Optional[bytes]] = {k: None for k in ids}

    def __iter__(self) -> KeysView[Union[str, bytes]]:
        return self.keys()

    def keys(self):
        # Overridden to prevent loading
        for k in self._data.keys():
            yield k

    def items(self):
        for k in self._data.keys():
            yield k, self[k]

    def __getitem__(self, __key) -> bytes:
        if __key not in self._data:
            raise KeyError(__key)
        if self._data[__key] is None:
            with open(os.path.join(self.__data_folder, __key), "rb") as f:
                self._data[__key] = f.read()
        return self._data[__key]

    def __len__(self) -> int:
        return len(self._data)


class TaskHandler:
    def __init__(self, request: ProcessRequest, agent_client: AgentStub):
        self._client: AgentStub = agent_client
        self.session_id: str = request.session_id
        self.task_id: str = request.task_id
        self.task_options: TaskOptions = TaskOptions.from_message(request.task_options)
        self.token: str = request.communication_token
        self.expected_results: List[str] = list(request.expected_output_keys)
        self.configuration: Configuration = request.configuration
        self.payload_id: str = request.payload_id
        self.data_folder: str = request.data_folder

        self._payload = None
        self.data_dependencies = LazyLoadDict(self.data_folder, list(request.data_dependencies))

    @property
    def payload(self) -> bytes:
        if self._payload is None:
            with open(os.path.join(self.data_folder, self.payload_id), "rb") as f:
                self._payload = f.read()
        return self._payload

    def submit_tasks(
        self,
        tasks: List[TaskDefinition],
        default_task_options: Optional[TaskOptions] = None,
        batch_size: Optional[int] = 100,
    ) -> List[Task]:
        """Submit tasks to the agent.

        Args:
            tasks: List of task definitions
            default_task_options: Default Task Options used if a task has its options not set
            batch_size: Batch size for submission
        """
        submitted_tasks: List[Task] = []
        for tasks_batch in batched(tasks, batch_size):
            task_creations = []

            for t in tasks_batch:
                task_creation = SubmitTasksRequest.TaskCreation(
                    expected_output_keys=t.expected_output_ids,
                    payload_id=t.payload_id,
                    data_dependencies=t.data_dependencies,
                )
                if t.options:
                    task_creation.task_options = t.options.to_message()
                task_creations.append(task_creation)

            request = SubmitTasksRequest(
                session_id=self.session_id,
                communication_token=self.token,
                task_creations=task_creations,
                task_options=(default_task_options.to_message() if default_task_options else None),
            )

            submitted_tasks.extend(
                Task(
                    id=t.task_id,
                    expected_output_ids=list(t.expected_output_ids),
                    data_dependencies=list(t.data_dependencies),
                    session_id=self.session_id,
                    payload_id=self.payload_id,
                )
                for t in self._client.SubmitTasks(request).task_infos
            )
        return submitted_tasks

    def send_results(self, results_data: Dict[str, Union[bytes, bytearray]]) -> None:
        """Send results.

        Args:
            results_data: A dictionary mapping each result ID to its data.
        """
        for result_id, result_data in results_data.items():
            with open(os.path.join(self.data_folder, result_id), "wb") as f:
                f.write(result_data)

        request = NotifyResultDataRequest(
            ids=[
                NotifyResultDataRequest.ResultIdentifier(
                    session_id=self.session_id, result_id=result_id
                )
                for result_id in results_data.keys()
            ],
            communication_token=self.token,
        )
        self._client.NotifyResultData(request)

    def create_results_metadata(
        self, result_names: List[str], batch_size: int = 100
    ) -> Dict[str, Result]:
        """
        Create the metadata of multiple results at once.
        Data have to be uploaded separately.

        Args:
            result_names: The names of the results to create.
            batch_size: Batch size for querying.

        Return:
            A dictionary mapping each result name to its result summary.
        """
        results = {}
        for result_names_batch in batched(result_names, batch_size):
            request = CreateResultsMetaDataRequest(
                results=[
                    CreateResultsMetaDataRequest.ResultCreate(name=result_name)
                    for result_name in result_names_batch
                ],
                session_id=self.session_id,
                communication_token=self.token,
            )
            response: CreateResultsMetaDataResponse = self._client.CreateResultsMetaData(request)
            for result_message in response.results:
                results[result_message.name] = Result.from_result_metadata(result_message)
        return results

    def create_results(
        self, results_data: Dict[str, bytes], batch_size: int = 1
    ) -> Dict[str, Result]:
        """Create one result with data included in the request.

        Args:
            results_data: A dictionary mapping the result names to their actual data.
            batch_size: Batch size for querying.

        Return:
            A dictionary mapping each result name to its corresponding result summary.
        """
        results = {}
        for results_ids_batch in batched(results_data.keys(), batch_size):
            request = CreateResultsRequest(
                results=[
                    CreateResultsRequest.ResultCreate(name=name, data=results_data[name])
                    for name in results_ids_batch
                ],
                session_id=self.session_id,
                communication_token=self.token,
            )
            response: CreateResultsResponse = self._client.CreateResults(request)
            for message in response.results:
                results[message.name] = Result.from_result_metadata(message)
        return results



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tasks_service.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import tasks_common_pb2 as tasks__common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13tasks_service.proto\x12\x19\x61rmonik.api.grpc.v1.tasks\x1a\x12tasks_common.proto2\xa6\x06\n\x05Tasks\x12h\n\tListTasks\x12+.armonik.api.grpc.v1.tasks.ListTasksRequest\x1a,.armonik.api.grpc.v1.tasks.ListTasksResponse\"\x00\x12x\n\x11ListTasksDetailed\x12+.armonik.api.grpc.v1.tasks.ListTasksRequest\x1a\x34.armonik.api.grpc.v1.tasks.ListTasksDetailedResponse\"\x00\x12\x62\n\x07GetTask\x12).armonik.api.grpc.v1.tasks.GetTaskRequest\x1a*.armonik.api.grpc.v1.tasks.GetTaskResponse\"\x00\x12n\n\x0b\x43\x61ncelTasks\x12-.armonik.api.grpc.v1.tasks.CancelTasksRequest\x1a..armonik.api.grpc.v1.tasks.CancelTasksResponse\"\x00\x12o\n\x0cGetResultIds\x12..armonik.api.grpc.v1.tasks.GetResultIdsRequest\x1a/.armonik.api.grpc.v1.tasks.GetResultIdsResponse\x12\x83\x01\n\x12\x43ountTasksByStatus\x12\x34.armonik.api.grpc.v1.tasks.CountTasksByStatusRequest\x1a\x35.armonik.api.grpc.v1.tasks.CountTasksByStatusResponse\"\x00\x12n\n\x0bSubmitTasks\x12-.armonik.api.grpc.v1.tasks.SubmitTasksRequest\x1a..armonik.api.grpc.v1.tasks.SubmitTasksResponse\"\x00\x42?\n!fr.aneo.armonik.api.grpc.v1.tasks\xaa\x02\x19\x41rmoniK.Api.gRPC.V1.Tasksb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tasks_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n!fr.aneo.armonik.api.grpc.v1.tasks\252\002\031ArmoniK.Api.gRPC.V1.Tasks'
  _globals['_TASKS']._serialized_start=71
  _globals['_TASKS']._serialized_end=877
# @@protoc_insertion_point(module_scope)



from typing import Union, Any, Type, TYPE_CHECKING, TypeVar

from google.protobuf.message import Message

"""
This file allows for better typing and autocompletion support
"""

SimpleOperator = int


class BasicMessage(Message):
    """
    Message type that requires a "field" kwarg
    """

    def __init__(self, *, field, **kwargs):
        self.field = field
        for k, v in kwargs.items():
            setattr(self, k, v)


if TYPE_CHECKING:
    BasicMessageType = Union[Type[BasicMessage], Type[Message]]
else:
    BasicMessageType = BasicMessage


class CombinationMessage(Message):
    """
    Message type for the ands and ors
    """

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


if TYPE_CHECKING:
    CombinationMessageType = Union[Type[CombinationMessage], Type[Message]]
    DisjunctionType = TypeVar("DisjunctionType", bound=CombinationMessageType)
    ConjunctionType = TypeVar("ConjunctionType", bound=CombinationMessageType)
else:
    CombinationMessageType = CombinationMessage
    DisjunctionType = CombinationMessageType
    ConjunctionType = CombinationMessageType


class InnerMessage(Message):
    """
    Message type for the value and operator
    """

    def __init__(self, *, value: Any, operator: SimpleOperator):
        self.value = value
        self.operator = operator


if TYPE_CHECKING:
    InnerMessageType = Union[Type[InnerMessage], Type[Message]]
else:
    InnerMessageType = InnerMessage



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: results_service.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import objects_pb2 as objects__pb2
from ..common import results_common_pb2 as results__common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15results_service.proto\x12\x1b\x61rmonik.api.grpc.v1.results\x1a\robjects.proto\x1a\x14results_common.proto2\x83\x0b\n\x07Results\x12r\n\x0bListResults\x12/.armonik.api.grpc.v1.results.ListResultsRequest\x1a\x30.armonik.api.grpc.v1.results.ListResultsResponse\"\x00\x12l\n\tGetResult\x12-.armonik.api.grpc.v1.results.GetResultRequest\x1a..armonik.api.grpc.v1.results.GetResultResponse\"\x00\x12y\n\x0eGetOwnerTaskId\x12\x32.armonik.api.grpc.v1.results.GetOwnerTaskIdRequest\x1a\x33.armonik.api.grpc.v1.results.GetOwnerTaskIdResponse\x12\x90\x01\n\x15\x43reateResultsMetaData\x12\x39.armonik.api.grpc.v1.results.CreateResultsMetaDataRequest\x1a:.armonik.api.grpc.v1.results.CreateResultsMetaDataResponse\"\x00\x12x\n\rCreateResults\x12\x31.armonik.api.grpc.v1.results.CreateResultsRequest\x1a\x32.armonik.api.grpc.v1.results.CreateResultsResponse\"\x00\x12\x83\x01\n\x10UploadResultData\x12\x34.armonik.api.grpc.v1.results.UploadResultDataRequest\x1a\x35.armonik.api.grpc.v1.results.UploadResultDataResponse\"\x00(\x01\x12\x89\x01\n\x12\x44ownloadResultData\x12\x36.armonik.api.grpc.v1.results.DownloadResultDataRequest\x1a\x37.armonik.api.grpc.v1.results.DownloadResultDataResponse\"\x00\x30\x01\x12\x84\x01\n\x11\x44\x65leteResultsData\x12\x35.armonik.api.grpc.v1.results.DeleteResultsDataRequest\x1a\x36.armonik.api.grpc.v1.results.DeleteResultsDataResponse\"\x00\x12\x84\x01\n\x11ImportResultsData\x12\x35.armonik.api.grpc.v1.results.ImportResultsDataRequest\x1a\x36.armonik.api.grpc.v1.results.ImportResultsDataResponse\"\x00\x12w\n\x17GetServiceConfiguration\x12\x1a.armonik.api.grpc.v1.Empty\x1a@.armonik.api.grpc.v1.results.ResultsServiceConfigurationResponse\x12u\n\x0cWatchResults\x12/.armonik.api.grpc.v1.results.WatchResultRequest\x1a\x30.armonik.api.grpc.v1.results.WatchResultResponse(\x01\x30\x01\x42\x43\n#fr.aneo.armonik.api.grpc.v1.results\xaa\x02\x1b\x41rmoniK.Api.gRPC.V1.Resultsb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'results_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n#fr.aneo.armonik.api.grpc.v1.results\252\002\033ArmoniK.Api.gRPC.V1.Results'
  _globals['_RESULTS']._serialized_start=92
  _globals['_RESULTS']._serialized_end=1503
# @@protoc_insertion_point(module_scope)



import datetime
import logging

from .conftest import all_rpc_called, rpc_called, get_client, data_folder
from armonik.common import TaskDefinition, TaskOptions
from armonik.worker import TaskHandler
from armonik.protogen.common.worker_common_pb2 import ProcessRequest
from armonik.protogen.common.objects_pb2 import Configuration


logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)


class TestTaskHandler:
    request = ProcessRequest(
        communication_token="token",
        session_id="session-id",
        task_id="task-id",
        expected_output_keys=["result-id"],
        payload_id="payload-id",
        data_dependencies=["dd-id"],
        data_folder=data_folder,
        configuration=Configuration(data_chunk_max_size=8000),
        task_options=TaskOptions(
            max_duration=datetime.timedelta(seconds=1), priority=1, max_retries=1
        ).to_message(),
    )

    def test_taskhandler_init(self):
        task_handler = TaskHandler(self.request, get_client("Agent"))

        assert task_handler.session_id == "session-id"
        assert task_handler.task_id == "task-id"
        assert task_handler.task_options == TaskOptions(
            max_duration=datetime.timedelta(seconds=1),
            priority=1,
            max_retries=1,
            partition_id="",
            application_name="",
            application_version="",
            application_namespace="",
            application_service="",
            engine_type="",
            options={},
        )
        assert task_handler.token == "token"
        assert task_handler.expected_results == ["result-id"]
        assert task_handler.configuration == Configuration(data_chunk_max_size=8000)
        assert task_handler.payload_id == "payload-id"
        assert task_handler.data_folder == data_folder
        assert task_handler.payload == "payload".encode()
        assert dict(task_handler.data_dependencies) == {"dd-id": "dd".encode()}

    def test_submit_tasks(self):
        task_handler = TaskHandler(self.request, get_client("Agent"))
        tasks = task_handler.submit_tasks(
            [
                TaskDefinition(
                    payload_id="payload-id",
                    expected_output_ids=["result-id"],
                    data_dependencies=[],
                )
            ]
        )

        assert rpc_called("Agent", "SubmitTasks")
        assert tasks is not None

    def test_send_results(self):
        task_handler = TaskHandler(self.request, get_client("Agent"))
        task_handler.send_results({"result-id": b"result data"})
        assert rpc_called("Agent", "NotifyResultData")

    def test_create_result_metadata(self):
        task_handler = TaskHandler(self.request, get_client("Agent"))
        results = task_handler.create_results_metadata(["result-name"])

        assert rpc_called("Agent", "CreateResultsMetaData")
        # TODO: Mock must be updated to return something and so that changes the following assertions
        assert results == {}

    def test_create_results(self):
        task_handler = TaskHandler(self.request, get_client("Agent"))
        results = task_handler.create_results({"result-name": b"test data"})

        assert rpc_called("Agent", "CreateResults")
        assert results == {}

    def test_lazy_load(self):
        handler = TaskHandler(self.request, None)
        assert handler.data_dependencies._data["dd-id"] is None
        assert len(handler.data_dependencies._data) == 1
        for _ in handler.data_dependencies.keys():
            pass
        assert handler.data_dependencies._data["dd-id"] is None
        for _ in handler.data_dependencies.items():
            pass
        assert handler.data_dependencies._data["dd-id"] == b"dd"

    def test_service_fully_implemented(self):
        assert all_rpc_called(
            "Agent", missings=["CreateTask", "GetCommonData", "GetDirectData", "GetResourceData"]
        )



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Dict, List, Optional

from deprecation import deprecated

from ..protogen.common.agent_common_pb2 import ResultMetaData
from ..protogen.common.applications_common_pb2 import ApplicationRaw
from ..protogen.common.tasks_common_pb2 import TaskDetailed, TaskSummary
from .filter import (
    FilterDescriptor,
    GenericTaskOptionsFilter,
    TaskFilter,
    TaskOptionFilter,
    SessionTaskOptionFilter,
    OutputFilter,
    SessionFilter,
    ResultFilter,
    PartitionFilter,
    ApplicationFilter,
)
from ..protogen.common.objects_pb2 import (
    Empty,
)
from ..protogen.common.objects_pb2 import (
    Output as WorkerOutput,
)
from ..protogen.common.objects_pb2 import (
    TaskOptions as RawTaskOptions,
)
from ..protogen.common.partitions_common_pb2 import PartitionRaw
from ..protogen.common.result_status_pb2 import ResultStatus as RawResultStatus
from ..protogen.common.results_common_pb2 import ResultRaw
from ..protogen.common.session_status_pb2 import SessionStatus as RawSessionStatus
from ..protogen.common.sessions_common_pb2 import SessionRaw
from .enumwrapper import ResultStatus, SessionStatus, TaskStatus
from .helpers import duration_to_timedelta, timedelta_to_duration, timestamp_to_datetime


@dataclass()
class TaskOptions:
    max_duration: timedelta
    priority: int
    max_retries: int
    partition_id: Optional[str] = None
    application_name: Optional[str] = None
    application_version: Optional[str] = None
    application_namespace: Optional[str] = None
    application_service: Optional[str] = None
    engine_type: Optional[str] = None
    options: Dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_message(cls, task_options):
        return cls(
            max_duration=duration_to_timedelta(task_options.max_duration),
            max_retries=task_options.max_retries,
            priority=task_options.priority,
            partition_id=task_options.partition_id,
            application_name=task_options.application_name,
            application_version=task_options.application_version,
            application_namespace=task_options.application_namespace,
            application_service=task_options.application_service,
            engine_type=task_options.engine_type,
            options={k: v for k, v in task_options.options.items()},
        )

    def to_message(self) -> RawTaskOptions:
        return RawTaskOptions(
            max_duration=timedelta_to_duration(self.max_duration),
            max_retries=self.max_retries,
            priority=self.priority,
            partition_id=self.partition_id,
            application_name=self.application_name,
            application_version=self.application_version,
            application_namespace=self.application_namespace,
            application_service=self.application_service,
            engine_type=self.engine_type,
            options=self.options,
        )


@dataclass()
class Output:
    error: Optional[str] = None

    @property
    def success(self) -> bool:
        return self.error is None

    def to_message(self):
        if self.error is None:
            return WorkerOutput(ok=Empty())
        return WorkerOutput(error=WorkerOutput.Error(details=self.error))


@dataclass()
class TaskDefinition:
    payload_id: str = ""
    expected_output_ids: List[str] = field(default_factory=list)
    data_dependencies: List[str] = field(default_factory=list)
    options: Optional[TaskOptions] = None

    def __post_init__(self):
        if len(self.expected_output_ids) <= 0:
            raise ValueError("expected_output_ids must be not be empty")


_task_filter = TaskFilter()


class TaskOptionDescriptor(FilterDescriptor):
    def __get__(self, instance, owner) -> GenericTaskOptionsFilter:
        # When called statically, returns the Task option filter
        if instance is None:
            return TaskOptionFilter() if issubclass(owner, Task) else SessionTaskOptionFilter()
        # Otherwise gets the object's hidden property
        return getattr(instance, self._name)


class OutputDescriptor(FilterDescriptor):
    def __get__(self, instance, owner) -> OutputFilter:
        # When called statically, returns the Task option filter
        if instance is None:
            return OutputFilter()
        # Otherwise gets the object's hidden property
        return getattr(instance, self._name)


class Task:
    id = FilterDescriptor(_task_filter)
    session_id = FilterDescriptor(_task_filter)
    owner_pod_id = FilterDescriptor(_task_filter)

    initial_task_id = FilterDescriptor(_task_filter)
    created_by = FilterDescriptor(_task_filter)
    parent_task_ids = FilterDescriptor(_task_filter)
    data_dependencies = FilterDescriptor(_task_filter)
    expected_output_ids = FilterDescriptor(_task_filter)
    retry_of_ids = FilterDescriptor(_task_filter)

    status = FilterDescriptor(_task_filter)
    status_message = FilterDescriptor(_task_filter)

    options = TaskOptionDescriptor(_task_filter)
    created_at = FilterDescriptor(_task_filter)
    submitted_at = FilterDescriptor(_task_filter)
    received_at = FilterDescriptor(_task_filter)
    acquired_at = FilterDescriptor(_task_filter)
    fetched_at = FilterDescriptor(_task_filter)
    started_at = FilterDescriptor(_task_filter)
    processed_at = FilterDescriptor(_task_filter)
    ended_at = FilterDescriptor(_task_filter)
    pod_ttl = FilterDescriptor(_task_filter)

    creation_to_end_duration = FilterDescriptor(_task_filter)
    processing_to_end_duration = FilterDescriptor(_task_filter)
    received_to_end_duration = FilterDescriptor(_task_filter)

    output = OutputDescriptor(_task_filter)

    pod_hostname = FilterDescriptor(_task_filter)
    payload_id = FilterDescriptor(_task_filter)

    def __init__(
        self,
        id: Optional[str] = None,
        session_id: Optional[str] = None,
        owner_pod_id: Optional[str] = None,
        initial_task_id: Optional[str] = None,
        created_by: Optional[str] = None,
        parent_task_ids: Optional[List[str]] = None,
        data_dependencies: Optional[List[str]] = None,
        expected_output_ids: Optional[List[str]] = None,
        retry_of_ids: Optional[List[str]] = None,
        status: TaskStatus = TaskStatus.UNSPECIFIED,
        status_message: Optional[str] = None,
        options: Optional[TaskOptions] = None,
        created_at: Optional[datetime] = None,
        submitted_at: Optional[datetime] = None,
        received_at: Optional[datetime] = None,
        acquired_at: Optional[datetime] = None,
        fetched_at: Optional[datetime] = None,
        started_at: Optional[datetime] = None,
        processed_at: Optional[datetime] = None,
        ended_at: Optional[datetime] = None,
        pod_ttl: Optional[datetime] = None,
        creation_to_end_duration: timedelta = timedelta(0),
        processing_to_end_duration: timedelta = timedelta(0),
        received_to_end_duration: timedelta = timedelta(0),
        output: Optional[Output] = None,
        pod_hostname: Optional[str] = None,
        payload_id: Optional[str] = None,
    ):
        self.id = id
        self.session_id = session_id
        self.owner_pod_id = owner_pod_id
        self.initial_task_id = initial_task_id
        self.created_by = created_by
        self.parent_task_ids = parent_task_ids
        self.count_parent_task_ids = (
            len(self.parent_task_ids) if self.parent_task_ids is not None else None
        )
        self.data_dependencies = data_dependencies
        self.count_data_dependencies = (
            len(self.data_dependencies) if self.data_dependencies is not None else None
        )
        self.expected_output_ids = expected_output_ids
        self.count_expected_output_ids = (
            len(self.expected_output_ids) if self.expected_output_ids is not None else None
        )
        self.retry_of_ids = retry_of_ids
        self.count_retry_of_ids = len(self.retry_of_ids) if self.retry_of_ids is not None else None
        self.status = status
        self.status_message = status_message
        self.options = options
        self.created_at = created_at
        self.submitted_at = submitted_at
        self.received_at = received_at
        self.acquired_at = acquired_at
        self.fetched_at = fetched_at
        self.started_at = started_at
        self.processed_at = processed_at
        self.ended_at = ended_at
        self.pod_ttl = pod_ttl
        self.creation_to_end_duration = creation_to_end_duration
        self.processing_to_end_duration = processing_to_end_duration
        self.received_to_end_duration = received_to_end_duration
        self.output = output
        self.pod_hostname = pod_hostname
        self.payload_id = payload_id

    def refresh(self, task_client) -> None:
        """Refresh the fields of this task object by using the given task client

        Args:
            task_client: ArmoniKTasks client
        """
        result: "Task" = task_client.get_task(self.id)
        self.session_id = result.session_id
        self.owner_pod_id = result.owner_pod_id

        self.initial_task_id = result.initial_task_id
        self.created_by = result.created_by
        self.parent_task_ids = result.parent_task_ids
        self.data_dependencies = result.data_dependencies
        self.expected_output_ids = result.expected_output_ids
        self.retry_of_ids = result.retry_of_ids

        self.status = result.status
        self.status_message = result.status_message

        self.options = result.options
        self.created_at = result.created_at
        self.submitted_at = result.submitted_at
        self.received_at = result.received_at
        self.acquired_at = result.acquired_at
        self.fetched_at = result.fetched_at
        self.started_at = result.started_at
        self.processed_at = result.processed_at
        self.ended_at = result.ended_at
        self.pod_ttl = result.pod_ttl

        self.creation_to_end_duration = result.creation_to_end_duration
        self.processing_to_end_duration = result.processing_to_end_duration
        self.received_to_end_duration = result.received_to_end_duration

        self.output = result.output

        self.pod_hostname = result.pod_hostname
        self.payload_id = result.payload_id
        self.is_init = True

    @classmethod
    def from_message(cls, task_raw: TaskDetailed) -> "Task":
        return cls(
            id=task_raw.id,
            session_id=task_raw.session_id,
            owner_pod_id=task_raw.owner_pod_id,
            initial_task_id=task_raw.initial_task_id,
            created_by=task_raw.created_by,
            parent_task_ids=list(task_raw.parent_task_ids),
            data_dependencies=list(task_raw.data_dependencies),
            expected_output_ids=list(task_raw.expected_output_ids),
            retry_of_ids=list(task_raw.retry_of_ids),
            status=TaskStatus(task_raw.status),
            status_message=task_raw.status_message,
            options=TaskOptions.from_message(task_raw.options),
            created_at=timestamp_to_datetime(task_raw.created_at),
            submitted_at=timestamp_to_datetime(task_raw.submitted_at),
            received_at=timestamp_to_datetime(task_raw.received_at),
            acquired_at=timestamp_to_datetime(task_raw.acquired_at),
            fetched_at=timestamp_to_datetime(task_raw.fetched_at),
            started_at=timestamp_to_datetime(task_raw.started_at),
            processed_at=timestamp_to_datetime(task_raw.processed_at),
            ended_at=timestamp_to_datetime(task_raw.ended_at),
            pod_ttl=timestamp_to_datetime(task_raw.pod_ttl),
            creation_to_end_duration=duration_to_timedelta(task_raw.creation_to_end_duration),
            processing_to_end_duration=duration_to_timedelta(task_raw.processing_to_end_duration),
            received_to_end_duration=duration_to_timedelta(task_raw.received_to_end_duration),
            output=Output(error=(task_raw.output.error if not task_raw.output.success else None)),
            pod_hostname=task_raw.pod_hostname,
            payload_id=task_raw.payload_id,
        )

    @staticmethod
    def from_summary(task_raw: TaskSummary) -> "Task":
        task = Task(
            id=task_raw.id,
            session_id=task_raw.session_id,
            owner_pod_id=task_raw.owner_pod_id,
            initial_task_id=task_raw.initial_task_id,
            created_by=task_raw.created_by,
            parent_task_ids=None,
            data_dependencies=None,
            expected_output_ids=None,
            retry_of_ids=None,
            status=TaskStatus(task_raw.status),
            status_message=task_raw.status_message,
            options=TaskOptions.from_message(task_raw.options),
            created_at=timestamp_to_datetime(task_raw.created_at),
            submitted_at=timestamp_to_datetime(task_raw.submitted_at),
            received_at=timestamp_to_datetime(task_raw.received_at),
            acquired_at=timestamp_to_datetime(task_raw.acquired_at),
            fetched_at=timestamp_to_datetime(task_raw.fetched_at),
            started_at=timestamp_to_datetime(task_raw.started_at),
            processed_at=timestamp_to_datetime(task_raw.processed_at),
            ended_at=timestamp_to_datetime(task_raw.ended_at),
            pod_ttl=timestamp_to_datetime(task_raw.pod_ttl),
            creation_to_end_duration=duration_to_timedelta(task_raw.creation_to_end_duration),
            processing_to_end_duration=duration_to_timedelta(task_raw.processing_to_end_duration),
            received_to_end_duration=duration_to_timedelta(task_raw.received_to_end_duration),
            output=Output(
                error=(
                    task_raw.error
                    if task_raw.error is not None and len(task_raw.error) > 0
                    else None
                )
            ),
            pod_hostname=task_raw.pod_hostname,
            payload_id=task_raw.payload_id,
        )
        task.count_parent_task_ids = task_raw.count_parent_task_ids
        task.count_data_dependencies = task_raw.count_data_dependencies
        task.count_expected_output_ids = task_raw.count_expected_output_ids
        task.count_retry_of_ids = task_raw.count_retry_of_ids
        return task

    def __eq__(self, other: "Task") -> bool:
        return (
            self.id == other.id
            and self.session_id == other.session_id
            and self.owner_pod_id == other.owner_pod_id
            and self.initial_task_id == other.initial_task_id
            and self.created_by == other.created_by
            and self.parent_task_ids == other.parent_task_ids
            and self.data_dependencies == other.data_dependencies
            and self.expected_output_ids == other.expected_output_ids
            and self.retry_of_ids == other.retry_of_ids
            and self.status == other.status
            and self.status_message == other.status_message
            and self.options == other.options
            and self.created_at == other.created_at
            and self.submitted_at == other.submitted_at
            and self.received_at == other.received_at
            and self.acquired_at == other.acquired_at
            and self.fetched_at == other.fetched_at
            and self.started_at == other.started_at
            and self.processed_at == other.processed_at
            and self.ended_at == other.ended_at
            and self.pod_ttl == other.pod_ttl
            and self.creation_to_end_duration == other.creation_to_end_duration
            and self.processing_to_end_duration == other.processing_to_end_duration
            and self.received_to_end_duration == other.received_to_end_duration
            and self.output == other.output
            and self.pod_hostname == other.pod_hostname
            and self.payload_id == other.payload_id
        )


@deprecated(deprecated_in="3.14.0", details="Use sessions, task and results client instead")
@dataclass
class ResultAvailability:
    errors: List[str] = field(default_factory=list)

    def is_available(self) -> bool:
        return len(self.errors) == 0


_sessionFilter = SessionFilter()


class Session:
    session_id = FilterDescriptor(_sessionFilter)
    status = FilterDescriptor(_sessionFilter)
    client_submission = FilterDescriptor(_sessionFilter)
    worker_submission = FilterDescriptor(_sessionFilter)
    partition_ids = FilterDescriptor(_sessionFilter)
    options = TaskOptionDescriptor(_sessionFilter)
    created_at = FilterDescriptor(_sessionFilter)
    cancelled_at = FilterDescriptor(_sessionFilter)
    closed_at = FilterDescriptor(_sessionFilter)
    purged_at = FilterDescriptor(_sessionFilter)
    deleted_at = FilterDescriptor(_sessionFilter)
    duration = FilterDescriptor(_sessionFilter)

    def __init__(
        self,
        session_id: Optional[str] = None,
        status: RawSessionStatus = SessionStatus.UNSPECIFIED,
        client_submission: Optional[bool] = None,
        worker_submission: Optional[bool] = None,
        partition_ids: Optional[List[str]] = None,
        options: Optional[TaskOptions] = None,
        created_at: Optional[datetime] = None,
        cancelled_at: Optional[datetime] = None,
        closed_at: Optional[datetime] = None,
        purged_at: Optional[datetime] = None,
        deleted_at: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
    ):
        self.session_id = session_id
        self.status = status
        self.client_submission = client_submission
        self.worker_submission = worker_submission
        self.partition_ids = partition_ids if partition_ids is not None else []
        self.options = options
        self.created_at = created_at
        self.cancelled_at = cancelled_at
        self.closed_at = closed_at
        self.purged_at = purged_at
        self.deleted_at = deleted_at
        self.duration = duration

    @classmethod
    def from_message(cls, session_raw: SessionRaw) -> "Session":
        return cls(
            session_id=session_raw.session_id,
            status=SessionStatus(session_raw.status),
            client_submission=session_raw.client_submission,
            worker_submission=session_raw.worker_submission,
            partition_ids=list(session_raw.partition_ids),
            options=TaskOptions.from_message(session_raw.options),
            created_at=timestamp_to_datetime(session_raw.created_at),
            cancelled_at=timestamp_to_datetime(session_raw.cancelled_at),
            closed_at=timestamp_to_datetime(session_raw.closed_at),
            purged_at=timestamp_to_datetime(session_raw.purged_at),
            deleted_at=timestamp_to_datetime(session_raw.deleted_at),
            duration=duration_to_timedelta(session_raw.duration),
        )

    def __eq__(self, other: "Session") -> bool:
        return (
            self.session_id == other.session_id
            and self.status == other.status
            and self.client_submission == other.client_submission
            and self.worker_submission == other.worker_submission
            and self.partition_ids == other.partition_ids
            and self.options == other.options
            and self.created_at == other.created_at
            and self.cancelled_at == other.cancelled_at
            and self.closed_at == other.closed_at
            and self.purged_at == other.purged_at
            and self.deleted_at == other.deleted_at
            and self.duration == other.duration
        )


_resultFilter = ResultFilter()


class Result:
    session_id = FilterDescriptor(_resultFilter)
    name = FilterDescriptor(_resultFilter)
    created_by = FilterDescriptor(_resultFilter)
    owner_task_id = FilterDescriptor(_resultFilter)
    status = FilterDescriptor(_resultFilter)
    created_at = FilterDescriptor(_resultFilter)
    completed_at = FilterDescriptor(_resultFilter)
    result_id = FilterDescriptor(_resultFilter)
    size = FilterDescriptor(_resultFilter)
    opaque_id = FilterDescriptor(_resultFilter)

    def __init__(
        self,
        session_id: Optional[str] = None,
        name: Optional[str] = None,
        created_by: Optional[str] = None,
        owner_task_id: Optional[str] = None,
        status: RawResultStatus = ResultStatus.UNSPECIFIED,
        created_at: Optional[datetime] = None,
        completed_at: Optional[datetime] = None,
        result_id: Optional[str] = None,
        size: Optional[int] = None,
        opaque_id: Optional[bytes] = None,
    ):
        self.session_id = session_id
        self.name = name
        self.created_by = created_by
        self.owner_task_id = owner_task_id
        self.status = status
        self.created_at = created_at
        self.completed_at = completed_at
        self.result_id = result_id
        self.size = size
        self.opaque_id = opaque_id

    @classmethod
    def from_message(cls, result_raw: ResultRaw) -> "Result":
        return cls(
            session_id=result_raw.session_id,
            name=result_raw.name,
            created_by=result_raw.created_by,
            owner_task_id=result_raw.owner_task_id,
            status=ResultStatus(result_raw.status),
            created_at=timestamp_to_datetime(result_raw.created_at),
            completed_at=timestamp_to_datetime(result_raw.completed_at),
            result_id=result_raw.result_id,
            size=result_raw.size,
            opaque_id=result_raw.opaque_id,
        )

    @classmethod
    def from_result_metadata(cls, result_metadata: ResultMetaData) -> "Result":
        return cls(
            session_id=result_metadata.session_id,
            name=result_metadata.name,
            status=result_metadata.status,
            created_at=timestamp_to_datetime(result_metadata.created_at),
            result_id=result_metadata.result_id,
        )

    def __eq__(self, other: "Result") -> bool:
        return (
            self.session_id == other.session_id
            and self.status == other.status
            and self.result_id == other.result_id
            and self.created_at == other.created_at
            and self.completed_at == other.completed_at
            and self.result_id == other.result_id
            and self.size == other.size
            and self.opaque_id == other.opaque_id
        )


_partitionFilter = PartitionFilter()


class Partition:
    id = FilterDescriptor(_partitionFilter)
    parent_partition_ids = FilterDescriptor(_partitionFilter)
    pod_reserved = FilterDescriptor(_partitionFilter)
    pod_max = FilterDescriptor(_partitionFilter)
    pod_configuration = FilterDescriptor(_partitionFilter)
    preemption_percentage = FilterDescriptor(_partitionFilter)
    priority = FilterDescriptor(_partitionFilter)

    def __init__(
        self,
        id: str,
        parent_partition_ids: List[str],
        pod_reserved: int,
        pod_max: int,
        pod_configuration: Dict[str, str],
        preemption_percentage: int,
        priority: int,
    ):
        self.id = id
        self.parent_partition_ids = parent_partition_ids
        self.pod_reserved = pod_reserved
        self.pod_max = pod_max
        self.pod_configuration = pod_configuration
        self.preemption_percentage = preemption_percentage
        self.priority = priority

    @classmethod
    def from_message(cls, partition_raw: PartitionRaw) -> "Partition":
        return cls(
            id=partition_raw.id,
            parent_partition_ids=list(partition_raw.parent_partition_ids),
            pod_reserved=partition_raw.pod_reserved,
            pod_max=partition_raw.pod_max,
            pod_configuration=dict(partition_raw.pod_configuration),
            preemption_percentage=partition_raw.preemption_percentage,
            priority=partition_raw.priority,
        )

    def __eq__(self, other: "Partition") -> bool:
        return (
            self.id == other.id
            and self.parent_partition_ids == other.parent_partition_ids
            and self.pod_reserved == other.pod_reserved
            and self.pod_max == other.pod_max
            and self.pod_configuration == other.pod_configuration
            and self.preemption_percentage == other.preemption_percentage
            and self.priority == other.priority
        )


_applicationFilter = ApplicationFilter()


class Application:
    name = FilterDescriptor(_applicationFilter)
    namespace = FilterDescriptor(_applicationFilter)
    service = FilterDescriptor(_applicationFilter)
    version = FilterDescriptor(_applicationFilter)

    def __init__(self, name: str, namespace: str, service: str, version: str):
        self.name = name
        self.namespace = namespace
        self.service = service
        self.version = version

    @classmethod
    def from_message(cls, application_raw: ApplicationRaw) -> "Application":
        return cls(
            name=application_raw.name,
            namespace=application_raw.namespace,
            service=application_raw.service,
            version=application_raw.version,
        )

    def __eq__(self, other: "Application") -> bool:
        return (
            self.name == other.name
            and self.namespace == other.namespace
            and self.service == other.service
            and self.version == other.version
        )
